# -*- coding: utf-8 -*-
"""
Import all necessary Matplotlib modules and set default options
To use this module, import * from it:

from NDimInv.plot_helper import *
"""
import os
import platform

# Latex support can be activated using an environment variable, otherwise the
# default settings are:
# - for windows: off
# - else: on
if('DD_USE_LATEX' in os.environ and os.environ['DD_USE_LATEX'] == '1'):
    use_latex = True
else:
    if platform.system() == "Windows":
        use_latex = False
    else:
        use_latex = True

import sys
already_loaded = 'matplotlib' in sys.modules

# just make sure we can access matplotlib as mpl
import matplotlib as mpl
if(not already_loaded):
    mpl.use('Agg')
version_strings = [x for x in mpl.__version__.split('.')]
version = [int(version_strings[0]), int(version_strings[1])]

# general settings
mpl.rcParams["lines.linewidth"] = 2.0
mpl.rcParams["lines.markeredgewidth"] = 3.0
mpl.rcParams["lines.markersize"] = 3.0
mpl.rcParams["font.size"] = 12

# hack-around for https://github.com/matplotlib/matplotlib/pull/2012
if(mpl.__version__ == '1.3.0'):
    pass
else:
    pass
mpl.rcParams['mathtext.default'] = 'regular'
# mpl.rcParams['font.weight'] = 400
# mpl.rcParams['font.family'] = 'Droid Sans'

if use_latex:
    mpl.rcParams['font.family'] = 'Open Sans'
    mpl.rcParams['text.usetex'] = True
    mpl.rcParams['text.latex.unicode'] = True

# sans-serif fonts
if(False):
    preamble = r'\usepackage{droidsans}\usepackage[T1]{fontenc} '
    preamble += r'\usepackage{sfmath} \renewcommand{\rmfamily}{\sffamily} '
    preamble += r'\renewcommand\familydefault{\sfdefault} '
    preamble += r'\usepackage{mathastext} '
    mpl.rc('text.latex', preamble=preamble)

import matplotlib.pyplot as plt
plt

import mpl_toolkits.axes_grid1 as axes_grid1
axes_grid1


def mpl_get_cb_bound_next_to_plot(ax):
    """
    Return the coordinates for a colorbar axes next to the provided axes
    object. Take into account the changes of the axes due to aspect ratio
    settings.

    Parts of this code are taken from the transforms.py file from matplotlib

    Important: Use only AFTER fig.subplots_adjust(...)

    Use as:
    =======
    cb_pos = mpl_get_cb_bound_next_to_plot(fig.axes[15])
    ax1 = fig.add_axes(cb_pos, frame_on=True)

    cmap = mpl.cm.jet_r
    norm = mpl.colors.Normalize(vmin=float(23), vmax=float(33))
    cb1 = mpl.colorbar.ColorbarBase(ax1, cmap=cmap, norm=norm,
                                    orientation='vertical')
    cb1.locator = mpl.ticker.FixedLocator([23,28,33])
    cb1.update_ticks()
    cb1.ax.artists.remove(cb1.outline)    # remove framei
    """
    position = ax.get_position()

    figW, figH = ax.get_figure().get_size_inches()
    fig_aspect = figH / figW
    box_aspect = ax.get_data_ratio()
    pb = position.frozen()
    pb1 = pb.shrunk_to_aspect(box_aspect, pb, fig_aspect).bounds

    ax_size = ax.get_position().bounds

    xdiff = (ax_size[2] - pb1[2]) / 2
    ydiff = (ax_size[3] - pb1[3]) / 2

    # the colorbar is set to 0.01 width
    sizes = [ax_size[0] + xdiff + ax_size[2] + 0.01,
             ax_size[1] + ydiff, 0.01, pb1[3]]

    return sizes


def mpl_get_cb_bound_below_plot(ax):
    """
    Return the coordinates for a colorbar axes below the provided axes object.
    Take into account the changes of the axes due to aspect ratio settings.

    Parts of this code are taken from the transforms.py file from matplotlib

    Important: Use only AFTER fig.subplots_adjust(...)

    Use as:
    =======
    """
    position = ax.get_position()

    figW, figH = ax.get_figure().get_size_inches()
    fig_aspect = figH / figW
    box_aspect = ax.get_data_ratio()
    pb = position.frozen()
    pb1 = pb.shrunk_to_aspect(box_aspect, pb, fig_aspect).bounds

    ax_size = ax.get_position().bounds

    # the colorbar is set to 0.01 width
    sizes = [ax_size[0], ax_size[1] - 0.14, pb1[2], 0.03]

    return sizes



# -*- coding: utf-8 -*-
"""
Regularization functions
"""
import numpy as np
import scipy.sparse as sparse


class DifferenceWeighting(object):
    """
    Implement a difference weighting scheme, e.g. to weight the regularization
    matrix with time differences

    See T. Günther (2004) - Inversion Methods and Resolution Analysis for the
    2D/3D Reconstruction of Resistivity Structures from DC Measurements. PhD
    Thesis, Bergbauakademie Freiburg, page 20
    """
    def __init__(self, data):
        self.data = data

    def get_C(self):
        r"""
        Create a diagonal matrix with
        :math:`C_{i,i} = \frac{1}{\sqrt(t_{i+1} - t_{i})}`
        """
        diff = np.diff(self.data)
        weighting_factors = 1 / np.sqrt(diff)
        # weighting_factors = 1 / (diff)

        C = np.diag(weighting_factors)
        C = sparse.csc_matrix(C)
        return C


class BaseRegularization(object):
    """
    Inherit all regularizations from this class
    """
    def __init__(self, decouple=None, outside_first_dim=None,
                 weighting_obj=None):
        """
        Parameters
        ----------
        decouple : [None or list/tuple] set all axes to 0 in order decouple
                   those axes. Start with zero.
        outside_first_dim : [None or list/tuple] this is only used for the
                            aggregation to all dimensions. If this object is
                            used to regularize dimensions other than 0, this
                            list decides which base-dimensions will be
                            regularized
        weighting_obj :
        """
        self.decouple = decouple
        self.outside_first_dim = outside_first_dim
        self.weighting_obj = weighting_obj

    def WtWm(self, parsize):
        Wm = self.Wm(parsize)
        WtWm = Wm.T.dot(Wm)
        return WtWm


class Damping(BaseRegularization):
    """
    Return a diagonal unit matrix as a damping matrix
    """
    def Wm(self, parsize):
        Wm = np.identity(parsize)
        # remove last row
        # Wm = Wm[0:-1, :]
        if(self.decouple is not None):
            for axis in self.decouple:
                Wm[:, axis] = 0
                Wm[axis, :] = 0
        return Wm


class SmoothingFirstOrder(BaseRegularization):
    """
    Implement a first order smoothing regularization object

    See Aster, page 98+
    """
    def Wm(self, parsize):
        """
        Return a first order smoothing matrix.

        Parameters
        ----------
        parsize : size of parameter vector m to be regularized

        """
        R = np.zeros((parsize - 1, parsize))
        for i in range(0, parsize - 1):
            R[i, i] = -1
            R[i, i + 1] = 1

        if(self.decouple is not None):
            for axis in self.decouple:
                R[:, axis] = 0
                R[axis, :] = 0

        if(self.weighting_obj is not None):
            # get weighting matrix C
            C = self.weighting_obj.get_C()
            R = C.dot(R)
        return R


class SmoothingSecondOrder(BaseRegularization):
    def Wm(self, parsize):
        """
        Return a second order smoothing matrix.

        Parameters
        ----------
        parsize : size of parameter vector m to be regularized

        """
        first_order = SmoothingFirstOrder()
        R1 = first_order.Wm(parsize)
        if(self.weighting_obj is not None):
            # get weighting matrix C
            C = self.weighting_obj.get_C()
            W = C.dot(R1)
        else:
            W = R1
        R = W.T.dot(W)
        RR = R.T.dot(R)
        return RR

        R = np.zeros((parsize - 2, parsize))
        for i in range(0, parsize - 2):
            R[i, i] = 1
            R[i, i + 1] = -2
            R[i, i + 2] = 1

        if(self.decouple is not None):
            for axis in self.decouple:
                R[:, axis] = 0
                R[axis, :] = 0

        return R



"""
Various helper functions
"""
import numpy as np


def print_visual_mul(W):
    """
    create a visual result of the multiplikation with m
    matrix-multiplication:
    row * column
    """
    for row in W:
        #print row
        expression = ''
        for par in range(0, row.size):
            if(row[par] != 0):
                if(row[par] < 0):
                    sign = ' - '
                else:
                    sign = ' + '

                nr = int(np.abs(row[par]))
                if(nr == 1):
                    nr = ''
                expression += '{0}{1}m{2}'.format(sign, nr, par + 1)
        expression = expression.strip()
        if(expression == ''):
            expression = '0'
        if(expression[0] == '+'):
            expression = expression[1:]
        print(expression)


def print_reg_mat(W):
    Wsquared = W.T.dot(W)
    print W
    print Wsquared
    print_visual_mul(Wsquared)



"""
Copyright 2014 Maximilian Weigand

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import itertools
import data_weighting


class ND_Data(object):
    """
    Data related functions
    """
    def __init__(self, model, extra_dims, data_weighting_key):
        self.D = None
        self.obj = model

        self.data_converter = None
        if data_weighting_key not in data_weighting.functions:
            raise Exception('data weighting scheme "{0}" is not known!'.format(
                data_weighting_key
            ))
        self.data_weighting_func = data_weighting.functions[data_weighting_key]

        # #### model side ####
        self.D_base_dims = self.obj.get_data_base_dimensions()
        self.D_base_size = [x[1][1] for x in self.D_base_dims.items()]
        self.extra_dims = extra_dims

        self.dimensions = self.D_base_dims.copy()
        for key, item in self.extra_dims.iteritems():
            self.dimensions[key + len(self.D_base_size)] = item

        self._base_length = None

        # extra dimensional mask
        self.extra_mask = None

    def _check_position_with_dimensions(self, extra_position):
        """
        Check if the provided position lies within the registered extra
        dimensional space

        Parameters
        ----------
        extra_position : list/tuple of extra dimensions. If no extra dimensions
                         are used, None has to be used.
        """
        # are there any extra dimensions?
        if(len(extra_position) == 0):
            if (len(self.extra_dims) == 0):
                return
            else:
                raise TypeError('We expect no extra dimensions')

        # correspond the number of extra dimensions?
        position_dims = len(extra_position) - 1
        if(position_dims not in self.extra_dims):
            raise TypeError(
                'The position requests more dimensions than were previously' +
                'registered')

        # are the requested indices located within the boundaries?
        for index, dimension in enumerate(extra_position):
            if(self.extra_dims[index][1] < (dimension + 1)):
                raise IOError('Requested extra dimensional position {0} ' +
                              'lies outside of registered size'.format(index))

    def D_iterator(self):
        """
        Generator for base_dimensional chunks of D, i.e. iterate over all
        extra dimensions of both D.

        Use as:

        >>> for d in self.D_iterator():
        >>>     self.Data.D[d]

        This function provides the slices in the same order as an oder='F'
        argument for .flatten().
        """
        all_indices = [range(0, x[1][1]) for x in self.extra_dims.iteritems()]
        extra_indices = itertools.product(*all_indices)

        # create slices for the base dimensions, corresponds to a list of ':'
        # for each base dimension in M or D
        D_base_indices = [slice(0, x[1][1]) for x in
                          self.D_base_dims.items()]

        # compute and save the forward responses
        for index in extra_indices:
            sd = tuple(D_base_indices + list(index))
            yield sd

    def WD(self):
        """
        Assemble the data weighting matrix using one of the available data
        weighting functions
        """
        WD = np.zeros_like(self.D)
        for slice_d in self.D_iterator():
            weightings = self.data_weighting_func(
                self.D[slice_d]
            )
            # weightings = data_weighting.get_weighting_one(basedata)
            # weightings = data_weighting.get_weighting_all_to_one(basedata)
            # weightings = data_weighting.get_weighting_rel_abs(
            #   self.D[slice_d])
            WD[slice_d] = weightings
        return WD

    @property
    def Wd(self):
        errors = self.WD().flatten(order='F')
        Wd = np.diag(errors)
        return Wd

    @property
    def Df(self):
        """
        Return a flattened version of D (data vector)
        """
        D = self.D
        if(self.extra_mask is not None):
            index = [slice(0, x) for x in self.D_base_size] + self.extra_mask
            D = D[tuple(index)]

        Df = D.flatten(order='F')
        return Df

    @property
    def base_length(self):
        """
        Return the length of the base dimensions in a flattened state
        """
        if(self._base_length is None):
            lengths = [x[1][1] for x in self.D_base_dims.iteritems()]
            base_length = 1
            for item in lengths:
                base_length *= item
            self._base_length = base_length

        return self._base_length

    def _allocate_D(self):
        """
        Allocate the array for self.D according to the dimensions and numbers
        in self.D_base_dims and self.extra_dims

        This function must only be called once after all dimensions were added.
        """
        if(self.D is not None):
            return

        # determine new shape of D
        new_shape = []
        for dimension in self.D_base_dims.keys():
            new_shape.append(self.D_base_dims[dimension][1])

        for dimension in self.extra_dims.keys():
            new_shape.append(self.extra_dims[dimension][1])

        self.D = np.zeros(new_shape)

    def add_data(self, data, data_format, extra):
        """
        Add data to the data space

        Parameters
        ----------
        data : ndarray corresponding to base dimensions
        data_format : data format as found in sip_convert
        extra : this tuple/list provides the position in the extra dimensional
                space for the data set. Use None if no extra dimensions are
                use.
        """
        if(self.data_converter is not None and
           self.obj.data_format is not None):
            data_converted = self.data_converter(
                data_format, self.obj.data_format, data)
        else:
            data_converted = data

        # allocate D if we didn't do it yet
        self._allocate_D()

        # check dimensions with self.D_extra_dims
        self._check_position_with_dimensions(extra)

        # update value nrs if appropriate
        for nr, dimension in enumerate(extra):
            new_dim = dimension + 1
            if(new_dim > self.extra_dims[nr][1]):
                self.extra_dims[nr][1] = new_dim

        # assign data
        # note: it is important to use a tuple to index a variable number of
        # dimensions
        # see:
        # file:///usr/share/doc/python-numpy-doc/html/user/basics.indexing.html
        index = [slice(0, x[1][1]) for x in self.D_base_dims.iteritems()]
        index += extra
        index = tuple(index)

        self.D[index] = data_converted



"""
Copyright 2014 Maximilian Weigand

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import itertools


class ND_Model(object):
    """
    For the model side we need to know some things about the underlying
    model. Here we use the Debye-Decomposition routines
    """
    def __init__(self, model, Data, extra_dims):
        self.extra_dims = extra_dims
        self.Data = Data
        self.obj = model
        # This function is used by the Iteration() class to plot spectra
        self.custom_plot_func = None

        """"
        Based upon the data base information, set the model base information:
        - determine tau-values (need frequencies for that)
        - determine starting parameters (need data for that)
        """
        self.M_base_dims = self.obj.get_model_base_dimensions()
        self.compute_starting_parameters()

        """
        for each dimension there can be multiple regularizations
        """
        # key == dimension
        self.regularizations = {}
        self.cached_Wms = {}

        # store function pointer for a steplength-selection function here
        self.steplength_selector = None

    def DM_iterator(self):
        """
        Generator for base_dimensional chunks of D and M, i.e. iterate over all
        extra dimensions of both D and M.

        Use as:

        >>> for d, m in self.DM_iterator():
        >>>     self.Data.D[d]
        >>>     self.M[m]

        This function provides the slices in the same order as an oder='F'
        argument for .flatten().
        """
        # TODO: should this be something like iteritems ?
        all_indices = [range(0, x[1][1]) for x in self.extra_dims.items()]
        extra_indices = itertools.product(*all_indices)

        # create slices for the base dimensions, corresponds to a list of ':'
        # for each base dimension in M or D
        M_base_indices = [slice(0, x[1][1]) for x in self.M_base_dims.items()]
        D_base_indices = [slice(0, x[1][1]) for x in
                          self.Data.D_base_dims.items()]

        # compute and save the forward responses
        for index in extra_indices:
            sm = tuple(M_base_indices + list(index))
            sd = tuple(D_base_indices + list(index))
            yield sd, sm

    def add_regularization(self, dimension, reg_object, lambda_object):
        """
        Parameters
        ----------
        dimension : number of dimension the new regularization belongs to
                    (starting with 0)
        reg_object : regularization object
        lam_method : method to choose the regularization parameter lambda for
                     this regularization from (fixed|quotient)
        lam_parameters : parameters for the lambda-selection process (e.g.
                         fixed lambda value)
        """
        if(dimension not in self.regularizations):
            self.regularizations[dimension] = []

        reg_set = (reg_object, lambda_object)
        self.regularizations[dimension].append(reg_set)

    def _get_steps(self, Msize):
        """
        TODO
        """
        steps = []
        for tmp_dim in range(0, len(Msize)):
            step = 1
            for i in range(0, tmp_dim):
                step *= Msize[i]
            steps.append(step)
        return steps

    def _permutate_indices(self, indices):
        if(not indices):
            return [0, ]

        dimind = indices[0]
        for item in indices[1:]:
            temp_ind = []
            for nr in item:
                for old_index in dimind:
                    if(type(old_index) == int):
                        old_index = [old_index, ]
                    temp_ind.append((old_index + [nr, ]))

            dimind = temp_ind
        return dimind

    def _get_offset(self, coordinates, dim_order, steps, dim):
        right_dims = []
        left_dims = []
        for nr2, i in enumerate(dim_order[1:]):
            if(i < dim):
                left_dims.append(coordinates[nr2])
            if(i > dim):
                right_dims.append(coordinates[nr2] * steps[i])
        offset = int(sum(right_dims))
        return offset

    def map_reg_matrix_to_global_Wm(self, dim, func, outside_first_dim):
        """
        Assemble the regularization matrix of one regularization function of
        the N dimensional problem. Depending on the function pointer that is
        provided, this can be either :math:`W_m^T \cdot W_m` or just
        :math:`W_m`

        TODO: Clean up, describe the procedure in more detail

        Parameters
        ----------
        dim : dimension to be regularized (int, starting with 0)
        func : function pointer for the regularization function, will be called
               as func(parametersize). This is usually a .Wm() function or
               WtWm() function.
        outside_first_dim : TODO
        """
        Msize = np.array(self.get_M_dimensions())

        # find all dimensions other than the requested one
        indices = list(set(range(0, Msize.size)) - set([dim, ]))

        # get new shape if we would swap the first with the regularized
        # dimension a'la Y = M.swapaxes(0, dim)
        new_size = Msize.copy()
        new_size[0], new_size[dim] = new_size[dim], new_size[0]

        # store changed order
        dim_order = range(0, len(Msize))
        dim_order[0] = dim
        dim_order[dim] = 0

        new_2d_shape = (np.prod(Msize[indices]), Msize[dim])
        Wm_small = func(new_2d_shape[1])
        size_x = new_2d_shape[0] * new_2d_shape[1]
        Wm_large = np.zeros((size_x, size_x))

        # # compute steps for each dimension
        # a step is the number of entries in the flattened version of the
        # parameter array which we have to skip in order to get the "neighbour"
        # of this regularization dimension
        steps = self._get_steps(Msize)

        indices = []
        for i in new_size[1:]:
            indices.append(range(0, i))

        # permutate indices
        dimind = self._permutate_indices(indices)

        last_offset = None  # store last offset to reset suboffset timer
        suboffset = -1

        # we collapse all dimensions expect for the regularization dimension,
        # and now have to get the regularization matrix for each of those
        # entries and map those to the global regularization matrix
        for nr in range(0, new_2d_shape[0]):
            width = new_2d_shape[1]
            # get coordinates of this iteration
            coordinates = dimind[nr]
            if(type(coordinates) == int):
                coordinates = [coordinates, ]

            if(dim > 0 and outside_first_dim is not None):
                if(coordinates[0] not in outside_first_dim):
                    continue

            # assign steps to the remaining dimensions, in the current order
            offset = 1

            offset = self._get_offset(coordinates, dim_order, steps, dim)
            if(last_offset is None):
                last_offset = offset

            if(last_offset != offset):
                last_offset = offset
                suboffset = -1

            suboffset += 1

            # 1) distance between entries: step
            # 2) offset: nr * pset
            start_x = offset + suboffset
            end_x = start_x + (width * steps[dim])

            # while the size of Wm is always the number of parameters, it can
            # vary for the y direction, depending on the type of matrix we are
            # aggregating. If we aggregate WtWm, we expect a NxN matrix. If we
            # aggregate only W, we expect an N-1xN matrix
            x = slice(start_x, end_x, steps[dim])
            y = self._get_y_slice(start_x, end_x, Wm_small.shape, dim, steps)
            Wm_large[y, x] = Wm_small
        return Wm_large

    def _get_y_slice(self, start_x, end_x, Wm_small_shape, dim, steps):
        nr_x, nr_y = Wm_small_shape
        if(nr_y == (nr_x - 1)):
            offset_y_slice = steps[dim]
        elif(nr_y == nr_x):
            offset_y_slice = 0
        else:
            print('Unexpected size of Wm!', Wm_small_shape)
            exit()

        y = slice(start_x, end_x - offset_y_slice, steps[dim])
        return y

    def retrieve_lams_and_WtWms(self, iteration):
        """
        Return a list of regularization matrices with corresponding lambda
        values.

        Parameters
        ----------
        iteration : the last iteration for which the new model update will be
                    computed
        """
        lambdas = []
        WtWms = []
        index = 0  # we need to know the absolute position of the lambda

        if(self.Data.extra_mask is None):
            dimensions = self.regularizations.keys()
        else:
            dimensions = [0, ]  # only base dimension

        for dimension in dimensions:
            for reg_set in self.regularizations[dimension]:
                # get the regularization matrix
                WtWm = self.map_reg_matrix_to_global_Wm(
                    dimension, func=reg_set[0].WtWm,
                    outside_first_dim=reg_set[0].outside_first_dim)
                # get the corresponding lambda value
                lambda_object = reg_set[1]
                lam = lambda_object.get_lambda(iteration, WtWm,
                                               lam_index=index)

                WtWms.append(WtWm)
                lambdas.append(lam)
                index += 1
        return lambdas, WtWms

    def get_M_dimensions(self):
        """
        get numbers of values for each dimension
        """
        sizes = [x[1][1] for x in self.M_base_dims.iteritems()]

        if(self.Data.extra_mask is None):
            sizes += [x[1][1] for x in self.extra_dims.iteritems()]
        return sizes

    def convert_to_M(self, m):
        """
        Given a flat vector m, return the multi-dimensional vecotor M shaped to
        the verious dimensions

        Parameters
        ----------
        m : flattened array
        """
        sizes = self.get_M_dimensions()
        M = m.reshape(sizes, order='F')
        return M

    def compute_starting_parameters(self):
        """
        Compute starting parameters for all spectra in the Data object
        """
        m0_list = []
        for d_slice, m_slice in self.DM_iterator():
            m0 = self.obj.estimate_starting_parameters(self.Data.D[d_slice])
            m0_list.append(m0)

        self.m0 = np.array(m0_list).flatten()

    def f(self, m):
        r"""
        Compute the forward response for the model parameters m

        Return the flattened response for ALL spectra in the form:

        :math:`\underline{f} = \begin{pmatrix}\underline{f}_1\\
        \underline{f}_2\\ \vdots\\ \underline{f}_n \end{pmatrix}`
        """
        response_list = []
        # nr of parameters (nr of tau values + 1 for rho0)
        step_size = self.M_base_dims[0][1]

        # loop over the spectra of all dimensions
        for index in range(0, m.size, step_size):
            pars = m[index: index + step_size]
            forward = self.obj.forward(pars).flatten(order='F')
            response_list.append(forward)
        response = np.array(response_list).flatten()
        return response

    def F(self, M):
        r"""Return the model responses in the same dimensionality as D, i.e.
        [base_dimensions, extra_dimensions].
        """
        F = np.zeros_like(self.Data.D)
        for d_slice, m_slice in self.DM_iterator():
            m_tiny = M[m_slice]
            F[d_slice] = self.obj.forward(m_tiny)
        return F

    def J(self, m):
        """
        Compute Jacobian for all dimensions
        """
        J_list = []
        # nr of parameters (nr of tau values + 1 for rho0)
        step_size = self.M_base_dims[0][1]
        for index in range(0, m.size, step_size):
            pars = m[index: index + step_size]
            J_temp = self.obj.Jacobian(pars)

            J_list.append(J_temp)

        # now build the block matrix
        block_height = J_list[0].shape[0]
        block_width = J_list[0].shape[1]

        matrix_width = len(J_list) * block_width
        matrix_height = len(J_list) * block_height

        J = np.zeros((matrix_height, matrix_width))

        for index in range(0, len(J_list)):
            start_x = index * block_width
            end_x = (index + 1) * block_width
            start_y = (index * block_height)
            end_y = (index + 1) * block_height
            J[start_y:end_y, start_x:end_x] = J_list[index]

        return J



"""
Copyright 2014 Maximilian Weigand

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
# from memory_profiler import *
import os
import logging
import scipy.sparse as sparse
import scipy.sparse.linalg as SL
import numpy as np
import json
import ND_Model
import ND_Data
from plot_helper import *
import reg_pars
import helper
helper
log = logging.getLogger(__name__)


class RMS_control(object):
    """Manage RMS definitions and names. Inherited by NDimInv

    Register rms values in the dict self.rms_types. For each rms value, add an
    list containing bools for each dimension of D. True denotes dimensions that
    are summed up for the rms value, and False denotes dimensions, along which
    rms values will be computed for each entry. It is advisable to only set one
    dimension to False.
    Missing entries in the list are automatically filled with True-values,
    i.e. they are summed up. Thus, an empy list denotes the rms of all data
    values.

    Examples:

    >>> self.rms_types['rms_all'] = []
    >>> self.rms_types['rms_re_im'] = [True, False]
    """
    def __init__(self):
        # default rms over all data points
        self.rms_types = {'rms_all': []}
        self.rms_names = {'rms_all': ['rms_all', ]}

    def add_rms(self, key, definition, names):
        """register a new rms value

        Parameters
        ----------
        key: internal key for this rms
        definition: rms definition, see class documentation
        names: a list of names corresponding to the rms values. The list has
               either only one entry for all rms values (numbers will be added
               automatically), or the length must match the number of resulting
               rms values.
        """
        if key in self.rms_types:
            print('WARNING: Duplicate RMS definition: {0}'.format(key))
        self.rms_types[key] = definition
        self.rms_names[key] = names

    def save_rms_definition(self, filename):
        # prepare a tuple containing both rms dicts
        rms_definition = (self.rms_types, self.rms_names)
        with open(filename, 'w') as fid:
            json.dump(rms_definition, fid)

    def load_rms_definition(self, filename):
        with open(filename, 'r') as fid:
            self.rms_types, self.rms_names = json.load(fid)


class RMS(object):
    """
    Augment the Iteration() class with RMS computation functions
    """

    @property
    def rms_values(self):
        r"""Compute RMS values based on the definitions in self.rms_values.
        Two RMS values are computed:

        .. math::

            RMS_{error} = \sqrt{\frac{1}{N} \sum_{i} \left( \frac{d_i -
                f_i}{\epsilon_i} \right)^2}

            RMS_{no~error} = \sqrt{\frac{1}{N} \sum_{i} \left(\frac{d_i -
                f_i}{1} \right)^2}

        RMS values without errors included are prepended with the string
        "_noerr", and with errors the string '_error' is added.
        """
        D = self.Data.D
        F = self.Model.F(self.Model.convert_to_M(self.m))
        diff = (D - F)

        WD = self.Data.WD()
        diff_err = diff * WD

        diff_sq = diff ** 2
        diff_err_sq = diff_err ** 2

        rms_values = {}
        for key, item in self.RMS.rms_types.iteritems():
            # determine which dimensions to sum up
            full_item = np.array(item + [True, ] * (len(D.shape) - len(item)))
            indices = np.where(full_item)[0]
            # remaining = np.where(~full_item)[0]

            for full_key, diff in zip((key + '_error', key + '_noerr'),
                                      (diff_err_sq, diff_sq)):
                rms_sum = np.sum(diff, axis=tuple(indices))
                N = np.prod([diff.shape[x] for x in indices])
                rms = np.atleast_1d(np.sqrt(rms_sum / N))
                rms_values[full_key] = rms
        return rms_values

    @property
    def old_rms_values(self):
        r"""
        Compute the rms values for the first part (usually real part or
        magnitude), the second part (usually imaginary part or phase), and both
        parts following the formula

        :math:`RMS=\sqrt{\frac{1}{N}\sum_i^N \frac{d_i - f_i(m)}{\epsilon_i}}`

        Returns
        -------
        rms_values : dict containing the following keys (
                     rms_part1_noerr, rms_part2_noerr, rms_both_noerr,
                     rms_part1_error, rms_part2_error, rms_both_error)

        """
        d = self.Data.Df
        f = self.Model.f(self.m)

        Wd = self.Data.Wd
        d_part1, d_part2 = self.split_data_into_parts(d)
        f_part1, f_part2 = self.split_data_into_parts(f)

        rms_values = []
        rms_values_ng = {}
        for di, fi, Wslice, name in (
            (d_part1, f_part1, slice(0, d_part1.size), 'part1'),
            (d_part2, f_part2, slice(d_part1.size, Wd.shape[0]), 'part2'),
                (d, f, slice(0, Wd.shape[0]), 'both')):
            diff = (di - fi)
            try:
                rms = np.sum(diff ** 2)
                rms /= float(di.size)
                rms = np.sqrt(rms)
            except FloatingPointError:
                rms = np.inf

            rms_values_ng['rms_' + name + '_noerr'] = rms

            try:
                rms_err = np.sum(Wd[Wslice, Wslice].dot(diff) ** 2)
                rms_err /= float(di.size)
                rms_err = np.sqrt(rms_err)
            except FloatingPointError:
                rms_err = np.inf

            rms_values_ng['rms_' + name + '_error'] = rms_err

            rms_values.append(rms)
        return rms_values_ng

    def split_data_into_parts(self, vector):
        """
        Split the provided vector into the two basic parts (re/im| mag/pha).
        This corresponds to a split in dimension 1
        """
        if(self.Data.extra_mask is None):
            Dsize = self.Data.D.shape
        else:
            Dsize = self.Data.D_base_size
        vec_temp = vector.reshape(Dsize, order='F')
        part1 = vec_temp[:, 0].flatten(order='F')
        part2 = vec_temp[:, 1].flatten(order='F')
        return part1, part2


class SearchSteplengthParFit(object):
    r"""
    Determine an optimal steplength parameter :math:`\alpha` by fitting a
    parabola through the results for :math:`\alpha \in [0, 0.5, 1]`.

    This procedure needs only two calculations of the model update
    (:math:`\alpha = [0.5, 1]`).

    By default we optimize the rms called 'rms_all_noerr'
    """
    def __init__(self, optimize='rms_all_noerr', optimize_index=0):
        """
        Parameters
        ----------
        optimize : which rms to optimize (default: all):
        """
        self.rms_key = optimize
        self.rms_index = optimize_index

    def get_steplength(self, it, par_update, ignore_all_err=False):
        """

        """

        old_rms = it.rms_values

        rms_values = []
        alpha_values = []

        rms_values.append(old_rms[self.rms_key][self.rms_index])
        alpha_values.append(0)

        for nr, test_alpha in enumerate((0.5, 1)):
            m_test = it.m + test_alpha * par_update
            it_test = it.copy()
            it_test.m = m_test
            try:
                test_rms = it_test.rms_values
                rms_values.append(test_rms[self.rms_key][self.rms_index])
                alpha_values.append(test_alpha)
            except FloatingPointError:
                pass
            except ArithmeticError:
                pass

        if(len(rms_values) != 3):
            print('Not all steplengths could be calculated. ' +
                  'Cannot fit parabola')
            return None

        # fit parabola
        x = np.array(alpha_values)
        y = np.array(rms_values)

        A = np.zeros((3, 3), dtype=np.float)
        A[:, 0] = x ** 2
        A[:, 1] = x
        A[:, 2] = 1
        a, b, c = np.linalg.solve(A, y)

        # compute minum minmum
        x_min = -b / (2 * a)

        # we need to make sure to lie in the range ]0, 1]
        if(x_min > 1):
            x_min = 1

        # use a default here, the inversion will probably end if we do not
        # improve the rms
        if(x_min <= 0):
            x_min = 0.1

        # debug plots
        if(False):
            fig, ax = plt.subplots(1, 1)
            ax.plot(x, y, '.')
            ax.set_xlabel('alpha')
            ax.set_label('rms')
            x_dense = np.linspace(0, 1, 30)
            ax.plot(x_dense, a * (x_dense ** 2) + b * x_dense + c, '-')
            ax.axvline(x_min)
            if('global_prefix' in it.Model.obj.settings):
                output_prefix = it.Model.obj.settings['global_prefix']
            else:
                output_prefix = ""
            filename = output_prefix + 'steplength_parbola_it{0}'.format(it.nr)
            fig.savefig(filename + '.png')
            fig.clf()
            plt.close(fig)
            del(fig)
        return x_min


class SearchSteplength(object):
    """
    Step length finder

    Check a list of fixed alpha values for the optimal value
    """
    def __init__(self, fixed_values=None, rms_key='rms_all_noerr',
                 rms_index=1):
        """
        Parameters
        ----------
        fixed_values : None for default values (0.1, 0.5, 1)
                       Provide a tuple of float values to set test values
        rms_key: RMS dict key to optimize
        rms_index: index for the provided RMS key
        """
        if(fixed_values is None):
            # 1e-5, 1e-4
            self.values = (1e-3, 1e-2, 0.1, 0.5, 1)
        else:
            self.values = fixed_values

        self.rms_key = rms_key
        self.rms_index = rms_index

    def get_steplength(self, it, par_update, ignore_all_err=False):
        r"""
        For a given Iteration and a new model update, find an optimal step
        length parameter :math:`\alpha`

        The variables self.rms_key/self.rms_index determine the RMS to use for
        the optimization.

        Note that we do not check if alpha == 0 yields optimal results, this is
        left to the stopping creteria.

        Return None if all steplengths lead to errors.

        :math:`m_{i+1} = m_i + \alpha \cdot \Delta m`

        Parameters
        ----------
        it : iteration object of last iteration
        par_update : model update for next iteration
        """
        old_rms = it.rms_values

        best_index = -1
        for nr, test_alpha in enumerate(self.values):
            m_test = it.m + test_alpha * par_update
            it_test = it.copy()
            it_test.m = m_test
            try:
                test_rms = it_test.rms_values

                if(best_index == -1):
                    best_index = nr
                    best_rms = test_rms[self.rms_key][self.rms_index]
                else:
                    if(old_rms[self.rms_key][self.rms_index] >
                       test_rms[self.rms_key][self.rms_index]):
                        best_index = nr
                        best_rms = test_rms[self.rms_key][self.rms_index]
            except FloatingPointError:
                pass
        if(best_index == -1):
            print('All tested steplengths caused exceptions. Somethings ' +
                  'really wrong here')
            return None

        print('Found an optimal steplength ({0}) with rms {1} '.format(
            self.values[nr], best_rms) + '(old rms: {0})'.format(
            old_rms[self.rms_key][self.rms_index]))
        best_value = self.values[best_index]
        return best_value


class InversionControl(object):
    """
    This class augments the NDimInv class with inversion control functions. The
    actual inversion code can be found in the Inversion class, which augments
    the Iteration class.
    """
    def __init__(self):
        super(InversionControl, self).__init__()
        self.iterations = []
        # which rms to use for stopping criteria
        self.stop_rms_key = 'rms_all_noerr'
        self.stop_rms_index = 0

    def get_initial_iteration(self):
        """Return a bare iteration object initialized with the starting model
        and all registered RMS values.

        Usually only the first iteration is initialized using this function.
        Afterwards, new iterations are based upon the last iteration.
        """
        it = Iteration(0, self.Data, self.Model, self.RMS, self.settings)

        # set starting model
        it.m = self.Model.m0
        it.f = self.Model.f(it.m)

        # set initial regularisation (lam0) values
        it.lams, _ = self.Model.retrieve_lams_and_WtWms(it)
        return it

    def start_inversion(self):
        """
        Initialize the inversion
        """
        it0 = self.get_initial_iteration()
        self.iterations.append(it0)

        # we store any additional inversion settings here
        # leave this for later use
        self.inversion_settings = {}

    def run_inversion(self):
        """
        Convenience wrapper which initializes and runs the full inversion with
        all iterations. This function also checks stopping criteria
        """
        log.info('Running inversion')
        if(self.iterations == []):
            self.start_inversion()
        # self.iterations[-1].plot()
        stop_now = False
        while(stop_now is False and
              not self.stop_before_next_iteration() and
              self.iterations[-1].nr < self.settings['max_iterations']):
            print('Iteration: {0}'.format(self.iterations[-1].nr + 1))

            new_iteration, stop_now = self.iterations[-1].next_iteration()

            # self.iterations[-1].plot()
            if(not stop_now):
                stop_now = self.check_stopping_criteria_before_update(
                    new_iteration)

            if(stop_now is False):
                self.iterations.append(new_iteration)

    def check_stopping_criteria_before_update(self, new_it):
        """
        Return True if one of the stopping criteria applies
        """

        # return if any NaN values are found in the new parameters
        if np.any(np.isnan(new_it.m)):
            return True

        # return if any value is below 1e-15
        if np.any(new_it.m[1:] < -15):
            return True
            raise Exception('values below 1e-15')

        rms_upd_eps = 1e-5  # min. requested rms change between iterations
        allowed_rms_im_increase_first_iteration = 1e2

        #
        nr = self.iterations[-1].nr
        old_rms = self.iterations[-1].rms_values[
            self.stop_rms_key][self.stop_rms_index]
        new_rms = new_it.rms_values[
            self.stop_rms_key][self.stop_rms_index]

        # if we are in the first iteration, then we allow a slight increase in
        # the imaginary RMS, but not above a certain threshold
        # TODO: Perhaps this threshold should be in RMS?
        if(new_rms > old_rms):
            if(nr == 0):
                increase = (new_rms - old_rms)
                if(increase > allowed_rms_im_increase_first_iteration):
                    print('First iteration RMS-IM increase lies above: ' +
                          '{0}'.format(
                              allowed_rms_im_increase_first_iteration))
                    return True
            else:
                # in all other cases: we do not allow an increase in rms
                print('RMS Increase')
                return True

        # stop of the rms increase does not lie above a certain threshold
        rms_diff = np.abs(new_rms - old_rms)
        if(rms_diff < rms_upd_eps):
            print('RMS update below threshold: {0} - {1} < {2}'.format(
                new_rms, old_rms, rms_upd_eps))
            return True

        return False

    def run_n_iterations(self, n):
        """
        Compute n iterations

        No stopping criteria are evaluated.
        """
        for i in range(0, n):
            new_iteration, stop_now = self.iterations[-1].next_iteration()
            self.iterations.append(new_iteration)
            # self.iterations[-1].plot()

    def stop_before_next_iteration(self):
        """
        Check stopping criteria
        """
        # for now most of the stopping criteria are evaluated after the
        # computating of the model update, but before the application of this
        # model update to the previous iteration
        return False


class Inversion(RMS):
    """
    This class augments the Iteration class with the actual inversion functions
    """

    def __init__(self):
        super(Inversion, self).__init__()
        self._update_dict = {}

    def next_iteration(self):
        """
        Compute next iteration and return the model update

        Returns
        -------
        new_iteration : Iteration object with the next iteration
        stop_now : if something goes wrong with the next iteration, return
                   False here an the inversion is stopped
        """
        new_iteration = Iteration(self.nr + 1, self.Data, self.Model,
                                  self.RMS, self.settings)

        lams, WtWms = self.Model.retrieve_lams_and_WtWms(self)
        new_iteration.lams = lams
        # debug: plot l-curve
        # import reg_pars
        # l_curve = reg_pars.Lcurve()
        # space Wtwm
        # l_curve._get_lambda(self, WtWms[0], 0)
        # # debug end

        update_m = self._model_update(lams, WtWms)

        # its (memory-wise) expensive to keep all details of the inversion
        # process. Delete them if requested
        retain_it_details = os.getenv('NDIMINV_RETAIN_DETAILS')
        if retain_it_details != '1':
            self._update_dict.clear()

        alpha = self.Model.steplength_selector.get_steplength(self, update_m)
        self._update_dict['alpha'] = alpha
        if(alpha is None):
            return None, True
        new_iteration.m = self.m + alpha * update_m
        new_iteration.f = self.Model.f(new_iteration.m)
        return new_iteration, False

    def _add_regularisations(self, A, b, lams, WtWms):
        r"""
        Add the regularization terms :math:`\lambda_i
        \underline{\unerline{W}}_{m,i}` to the matrix A and to vector b.
        """
        # add regularizations
        for lam, WtWm in zip(lams, WtWms):
            WtWm_sparse = sparse.csc_matrix(WtWm)

            if(type(lam) is not int and not isinstance(lam, float)):
                A = A + lam.dot(WtWm_sparse)
                b = b - lam.dot(WtWm_sparse).dot(self.m)
            else:
                A = A + lam * WtWm_sparse
                b = b - lam * WtWm_sparse.dot(self.m)
        return A, b

    def _select_solver(self):
        """
        Based on the environment variable DD_SOLVER, return the function which
        solves the system of linear equations A*x + b using the call

        >>> x = solve_func(A, b)

        Possible choices for DD_SOLVER:

            *  std: use scipy.sparse.linalg.spsolve
            *  cg: use scipy.sparse.linalg.cg (sparse conjugate gradient)
        """
        if('DD_SOLVER' in os.environ):
            solver_id = os.environ['DD_SOLVER']
            if(solver_id == 'std'):
                solve_func = SL.spsolve
            elif(solver_id == 'cg'):
                solve_func = SL.cg
            else:
                print('ERROR: Solver not known: {0}'.format(solver_id))
                exit()
        else:
            solve_func = SL.spsolve
        # print('Using function {0} to solve system of lin. equations'.format(
        #     solve_func))
        return solve_func

    def _prepare_mode_update(self, lams, WtWms):
        """

        """
        diff = self.Data.Df - self.f
        J = sparse.csc_matrix(self.Model.J(self.m))
        Wd = sparse.csc_matrix(self.Data.Wd)
        self._update_dict['diff'] = diff
        self._update_dict['J'] = J
        self._update_dict['Wd'] = Wd
        self._update_dict['A_without_reg'] = J.T.dot(Wd.T.dot(Wd.dot(J)))
        self._update_dict['b_without_reg'] = J.T.dot(Wd.T.dot(Wd.dot(diff)))
        A, b = self._add_regularisations(self._update_dict['A_without_reg'],
                                         self._update_dict['b_without_reg'],
                                         lams, WtWms)
        self._update_dict['A'] = A
        self._update_dict['b'] = b
        self._update_dict['lams'] = lams
        self._update_dict['WtWms'] = WtWms

    def _model_update(self, lams, WtWms):
        """
        Compute the actual model update with the regularisations provided

        Parameters
        ----------
        last_it : Iteration object on which this update is based on
        lams : list/tuple of lambdas
        WtWms : list/tuple of regularization matrices :math:`W^T \cdot W`

        Returns
        -------
        model_update : model update vector
        """
        self._prepare_mode_update(lams, WtWms)

        # gather components of update formula
        solve_func = self._select_solver()

        # solve the system of linear equations A x = b
        update_m = solve_func(self._update_dict['A'],
                              self._update_dict['b'])

        # save A and b to file
        # import pickle
        # with open('A.dat', 'wb') as outfile:
        #     pickle.dump(A, outfile, pickle.HIGHEST_PROTOCOL)
        # np.savetxt('b.dat', b)

        # sometimes we get a list, in this instance take only the first element
        # see documentation for the solver functions
        if(isinstance(update_m, tuple)):
            update_m = update_m[0]
        self._update_dict['update'] = update_m
        return update_m


class Iteration(Inversion):
    """
    This class holds all information of one iteration
    """
    def __init__(self, nr, Data, Model, RMS, settings):
        """
        Parameters
        ----------
        nr : iteration nr
        Data : Data object
        Model : Model object
        rms_types : dict defining the rms types to be computed
        settings: settings-dict as used by the NDimInv object
        """
        super(Iteration, self).__init__()

        self.RMS = RMS
        self.nr = nr
        self.Data = Data
        self.Model = Model
        self.settings = settings

        # the following parameters will be computed during the inversion
        # process
        self.m = None
        self.f = None
        self.lams = None
        self.statpars = None

    def plot(self, filename='iteration', **kwargs):
        output_filename = "plot_"
        if('global_prefix' in self.Model.obj.settings):
            output_filename += self.Model.obj.settings['global_prefix']
        else:
            pass

        output_filename += filename + '{0:04}.png'.format(self.nr)
        if(self.Model.custom_plot_func is None):
            self._plot_default(output_filename)
        else:
            self.Model.custom_plot_func.plot(self, output_filename, **kwargs)

    def _plot_default(self, filename):
        """
        Default plot routine for complex resistivity data
        """
        response = self.Model.f(self.m)
        d = self.Data.Df
        nr_f = len(self.Data.obj.frequencies)
        nr_spectra = response.size / 2 / nr_f

        # split data
        slices = []
        resp = []
        for i in range(0, d.size, nr_f):
            slices.append(d[i:i + nr_f])
            resp.append(response[i:i + nr_f])

        size_y = 2 * nr_spectra
        fig, axes = plt.subplots(nr_spectra, 2, figsize=(7, size_y))
        axes = np.atleast_2d(axes)

        for nr, i in enumerate(range(0, nr_spectra * 2, 2)):
            # part 1
            ax = axes[nr, 0]
            ax.semilogx(self.Data.obj.frequencies, slices[i], '.-',
                        label='data')
            ax.semilogx(self.Data.obj.frequencies, resp[i], '.-', c='r',
                        label='fit')
            ax.legend(loc='best')

            # part 2
            ax = axes[nr, 1]
            ax.semilogx(self.Data.obj.frequencies, slices[i + 1], '.-')
            ax.semilogx(self.Data.obj.frequencies, resp[i + 1], '.-', c='r')

        fig.suptitle(self.nr)
        fig.savefig(filename, dpi=150)
        fig.clf()
        plt.close(fig)
        del(fig)

    def copy(self):
        """
        Return a copy of this instance. This copy is not a full copy,
        self.Data/self.Model will only be copied by reference. Thus if you
        change them here they will be changed everywhere.
        """
        it_copy = Iteration(self.nr,
                            self.Data,
                            self.Model,
                            self.RMS,
                            self.settings)

        # copy all variables
        it_copy.m = self.m
        it_copy.lams = self.lams
        it_copy.f = self.f

        return it_copy

    @property
    def stat_pars(self):
        """
        Aggregate statistical parameters for this iteration and return a
        dictionary
        """
        self.statpars = {}
        # loop over the m instances (i.e. the base dimensions)
        parsize = self.Model.M_base_dims[0][1]
        for nr, index in enumerate(range(0, self.m.size, parsize)):
            one_parset = self.m[index: index + parsize]
            single_par_stats = self.Model.obj.compute_par_stats(one_parset)
            # now sort into self.statpars
            for key, item in single_par_stats.iteritems():
                if(key not in self.statpars):
                    self.statpars[key] = []
                self.statpars[key].append(item)

        return self.statpars

    def plot_reg_strengths1(self, ax1, ax2):
        r"""
        Plot :math:`\left[\underline{\underline{J}}^T
        \underline{\underline{W}}_d^T \underline{\underline{W}}_d
        (\underline{d} - \underline{f}(\underline{m})) \right]^{-1}`
        """
        diff = self.Data.Df - self.f
        J = sparse.csc_matrix(self.Model.J(self.m))

        # compute update
        # base
        Wd = sparse.csc_matrix(self.Data.Wd)

        reg_strength1 = J.T.dot(Wd.T.dot(Wd.dot(diff)))
        # split into rho0 and chargeabilties
        size_m = self.Model.obj.tau.size

        rho0 = reg_strength1.copy()
        # mask chargeabilities
        for i in range(1, rho0.size, size_m + 1):
            rho0[i:i + size_m] = np.nan
        ax1.plot(rho0, '.-', color='k')
        ax1.set_xlim([0, len(reg_strength1)])
        ax1.grid(True)
        ax1.set_title('$rho_0$ - mean: {0} max: {1}'.format(
            np.mean(rho0), np.nanmax(rho0)))

        for i in range(0, rho0.size, size_m + 1):
            ax1.axvline(x=i, color='k', linestyle='dashed', linewidth=0.5)

        if('DD_USE_LATEX' in os.environ and
           os.environ['DD_USE_LATEX'] == '1'):
            ylabel_base = r'\left[\underline{\underline{J}}^T '
            ylabel_base += r'\underline{\underline{W}}_d^T '
            ylabel_base += r'\underline{\underline{W}}_d (\underline{d} - '
            ylabel_base += r'\underline{f}(\underline{m})) \right]^{-1}'
        else:
            ylabel_base = '[J^t W_d^T W_d (d - f(m))]^{-1}'
        ylabel_base = '$' + ylabel_base + '$'
        ax1.set_ylabel(ylabel_base)

        m = reg_strength1
        # mask rho0 values
        for i in range(0, m.size, size_m + 1):
            m[i] = np.nan

        ax2.plot(m, '.-', color='k')
        # now mark the start of each parameter set
        for i in range(0, m.size, size_m + 1):
            ax2.axvline(x=i, color='k', linestyle='dashed', linewidth=0.5)
        ax2.set_ylabel(ylabel_base)

        ax2.set_xlim([0, len(reg_strength1)])
        ax2.grid(True)
        ax2.set_title('$m_i$ - mean: {0} max: {1}'.format(
            m.mean(), np.nanmax(m)))

    # @profile
    def plot_reg_strengths(self):
        r"""
        For each registered regularisation, plot :math:`\lambda_i \underline{
        \underline{W}}^T \underline{\underline{W}} \cdot \underline{m}` vs
        :math:`\underline{m}`
        """
        if('global_prefix' in self.Model.obj.settings):
            output_prefix = self.Model.obj.settings['global_prefix']
        else:
            output_prefix = ""

        outfile = output_prefix + 'reg_strength'
        lams, WtWms = self.Model.retrieve_lams_and_WtWms(self)
        fig, axes = plt.subplots(len(lams) + 2, 1,
                                 figsize=(7, len(lams) * 2.5 + 4))
        size_m = self.Model.obj.tau.size
        reg_index = 0
        for lam, WtWm in zip(lams, WtWms):
            ax = np.atleast_1d(axes)[reg_index]
            reg_strength = WtWm.dot(self.m)
            mean_strength = np.mean(reg_strength)
            if(isinstance(lam, (int, float))):
                reg_strength_lam = lam * reg_strength
                lam_value = lam
            else:
                reg_strength_lam = lam.dot(reg_strength)
                lam_value = 'indiv'

            ax.set_title('lam: {0}, mean wo lambda: {1}, max: {2}'.format(
                lam_value, mean_strength, np.max(reg_strength_lam)))
            ax.plot(reg_strength_lam, '.-', color='k')
            # now mark the start of each parameter set
            for i in range(0, self.m.size, size_m + 1):
                ax.axvline(x=i, color='k', linestyle='dashed', linewidth=0.5)
            ax.set_xlim([0, len(reg_strength)])
            ax.set_xlabel('Parameter number')
            if('DD_USE_LATEX' in os.environ and
               os.environ['DD_USE_LATEX'] == '1'):
                ylabel = r''.join((
                    r'$\lambda \cdot \underline{\underline{W}}^T ',
                    r'\underline{\underline{W}} \underline{m}$'
                ))
            else:
                ylabel = '$lam W^t W m$'

            ax.set_ylabel(ylabel)
            ax.grid(True)
            reg_index += 1

        self.plot_reg_strengths1(axes[-2], axes[-1])
        fig.tight_layout()
        fig.savefig(outfile + '.png')
        fig.clf()
        plt.close(fig)
        del(fig)

        # save plot data
        np.savetxt(outfile + '.dat', reg_strength_lam)

    def plot_lcurve(self):
        """
        plot the L-curve after hansen
        """

        print('Plotting l-curve for iteration {0}'.format(self.nr))
        if('global_prefix' in self.Model.obj.settings):
            output_prefix = self.Model.obj.settings['global_prefix']
        else:
            output_prefix = ""

        LCURVE = reg_pars.Lcurve()
        lams, WtWms = self.Model.retrieve_lams_and_WtWms(self)

        for lam_index in range(0, len(lams)):
            prefix = output_prefix + 'lam-nr_{0}_'.format(lam_index)
            LCURVE.plot_lcurve(self, WtWms, lams, lam_index, prefix)
            lam_index += 1


class NDimInv(InversionControl):
    """
    N-dimensional model inversion for SIP-Spectra

    Data parameters
    ===============

    Data is stored in a numpy-array self.D

    Information on additional dimensions (extra dimensions) can be found in the
    dict self.D_extra_dims

    Model parameters are stored in self.M

    Usage examples
    ==============
    Implement the following use cases:

    1) Invert one spectrum (i.e. no extra dimensions)
    2) Time regularisation: invert 5 spectra for different time steps
    3) Spatial regularisation: invert 9 spectra for one time step at different
       locations
    4) Invert 9 x 5 spectra for 9 locations at 5 time steps

    TODO:
    =====
    * Find a way to check if all data was added or if we have 'white' spots in
      self.D

    """
    def __init__(self, model, settings):
        # call __init__ function of the InversionControl class
        super(NDimInv, self).__init__()

        self.Data = None
        self.Model = None
        self.settings = settings
        self._check_settings()
        self.model = model
        self.RMS = RMS_control()

        # #### extra dimensions #####
        # extra dimensions both are associated with the data and the model
        # space
        self.extra_dims = {}

    def _check_settings(self):
        """
        Check if all required settings were made
        """
        required_settings = ('max_iterations', )
        for requirement in required_settings:
            if(requirement not in self.settings):
                print('NDimInv: Missing required setting {0}'.format(
                    requirement))
                exit()

    def finalize_dimensions(self):
        """
        After all dimensions were added, create the Data space
        """
        data_weighting_function = self.settings.get(
            'data_weighting',
            None,
        )

        self.Data = ND_Data.ND_Data(
            self.model,
            self.extra_dims,
            data_weighting_function,
        )

    def set_custom_plot_func(self, plot_obj):
        """
        Each iteration can plot the spectrum. We can provide a custom plot
        function to plot more detailed information
        """
        self.Model.custom_plot_func = plot_obj

    def update_model(self):
        """
        Create the model object
        """
        self.Model = ND_Model.ND_Model(self.model, self.Data,
                                       self.extra_dims)

    def _get_nr_dims(self, dim_dict):
        """
        Return number of dimensions for a provided dictionary

        Parameters
        ----------
        dim_dict: dict object such as self.D_base_dims or self.extra_dims
        """
        keys = dim_dict.keys()
        return len(keys)

    @property
    def nr_base_dims(self):
        """
        Return number of base dimensions
        """
        return self._get_nr_dims(self.D_base_dims)

    @property
    def nr_extra_dims(self):
        """
        Return number of extra dimensions
        """
        return self._get_nr_dims(self.extra_dims)

    def add_new_dimension(self, description, size):
        """
        Add a new extra dimension to the inversion.

        Extra dimensions are added both to data and model space.

        Parameters
        ----------
        description : short string describing this dimension. Shouldn't contain
                      spaces
        size : number of entries on this dimension (e.g. how many time
                      steps or how many locations)
        """
        # get largest dimension number
        # use this as the index of the new dimension
        next_dim_nr = self.nr_extra_dims
        self.extra_dims[next_dim_nr] = [description, size]

    def _print_data_dimension_information(self, dim_dict):
        """
        Print information regarding a dimension dict (e.g. base or extra)
        """
        for key in dim_dict.keys():
            print('{0} {1} [{2} values]'.format(key, dim_dict[key][0],
                                                dim_dict[key][1]))

    def overview_data(self):
        """
        Print information regarding the data space self.D
        """
        print('\n\nDimension Information\n\n')
        print('base dimensions')
        print('---------------')
        self._print_data_dimension_information(self.Data.D_base_dims)
        if(self.Model is not None):
            self._print_data_dimension_information(self.Model.M_base_dims)

        print('')
        print('extra dimensions')
        print('---------------')
        self._print_data_dimension_information(self.extra_dims)



"""
Template class for models

Inherit from it when you implement fit models
"""
import abc


class model_template(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self, settings):
        """
        Parameters
        ----------
        settings : dict containing the settings for the forward model. These
                   settings are mode dependent, but usually include a parameter
                   (e.g. 'x') which serves as the independent variable
                   associated with the model parameters.

        """
        self.settings = settings
        # set this to None if no data conversion is to done
        self.data_format = None

    @abc.abstractmethod
    def estimate_starting_parameters(self, base_data):
        """Given a data set of base data dimensions, return an initial guess
        (starting parameters) for the inversion.

        Parameters
        ----------
        base_data : input data with base dimensions

        Returns
        -------
        initial_pars : initial parameters, size model base dimensions

        """
        print('need a function ' +
              '"estimate_starting_parameters(self, base_data)"')
        exit()

    @abc.abstractmethod
    def forward(self, pars):
        """Return the forward response in base dimensions
        """
        print('"forward" function not implemented')
        exit()

    @abc.abstractmethod
    def Jacobian(self, pars):
        r"""Return the Jacobian corresponding to the forward response. The
        Jacobian has the dimensions :math:`B \times D \times M`

        TODO: Check the return dimensions
        """
        print('"Jacobian" function not implemented')
        exit()

    @abc.abstractmethod
    def get_data_base_size(self):
        """Usually you do not need to modify this.
        """
        size = sum([x[1][1] for x in
                    self.get_data_base_dimensions().iteritems()])
        return size

    @abc.abstractmethod
    def get_data_base_dimensions(self):
        """
        Returns
        -------
        Return a dict with a description of the data base dimensions. In this
        case we have frequencies and re/im data

        In the example down below, the frequencies will be set using the
        settings dict supplied to the __init__ function. Therefore the number of
        frequencies must be added dynamically here. Do not leave any values as
        None!
        """
        print('base data dimensions not set')
        exit()
        D_base_dims = {0: ['frequency', None],
                       1: ['rre_rmim', 2]
                       }
        return D_base_dims

    @abc.abstractmethod
    def get_model_base_dimensions(self):
        """Return a dict with a description of the model base dimensions. In
        this case we have one dimension: the DD parameters (rho0, mi) where m_i
        denotes all chargeability values corresponding to the relaxation times.
        """
        M_base_dims = {0: ['rho0_mi', self.tau.size + 1]}
        return M_base_dims

    @abc.abstractmethod
    def compute_par_stats(self, pars):
        r"""For a given parameter set (i.e. a fit result), compute relevant
        statistical values such as :math:`m_{tot}`, :math:`m_{tot}^n`,
        :math:`\tau_{50}`, :math:`\tau_{mean}`, :math:`\tau_{peak}`

        This is the way to compute any secondary results based on the fit
        results.

        Store in self.stat_pars = dict()
        """
        self.stat_pars = {}
        print('function "compute_par_stats" is not implemented')
        exit()
        return self.stat_pars



"""
Collection of helper functions which generate weightings for various base data
sets
"""
import numpy as np


def get_weighting_re_vs_im(base_data):
    r"""Return a vector of weighting factors that compensate the mean of real
    and imaginary parts:

    :math:`W_{im} = \frac{\overline{|re|}}{\overline{|im|}}`

    We assume that base_data is of shape (N, 2) with (N, 0) the real parts, (N,
    1) the imaginary parts
    """
    work_data = base_data.flatten(order='F')

    center = int(work_data.size / 2)
    re = work_data[0: center]
    im = work_data[center:]

    re_mean = np.mean(np.abs(re))
    im_mean = np.mean(np.abs(im))

    factor_im = re_mean / im_mean
    errors = np.ones(work_data.shape)
    errors[center:] *= factor_im

    errors = errors.reshape(base_data.shape, order='F')
    return errors


def get_weighting_im_to_avg_re(base_data):
    r"""Weigh imaginary parts to the mean of the real part

    :math:`W_{im} = \frac{\overline{|re|}}{\overline{|im|}}`

    We assume that base_data is of shape (N, 2) with (N, 0) the real parts, (N,
    1) the imaginary parts
    """
    work_data = base_data.flatten(order='F')

    center = int(work_data.size / 2)
    re = work_data[0: center]
    im = work_data[center:]

    re_mean = np.mean(np.abs(re))

    factor_im = re_mean / im
    errors = np.ones(work_data.shape)
    errors[center:] *= factor_im

    errors = errors.reshape(base_data.shape, order='F')
    return errors


def get_weighting_im_to_avg_re_error(base_data):
    r"""Weigh imaginary parts to the mean of the real part

    :math:`W_{im} = \frac{\overline{|re|}}{\overline{|im|}}`

    We assume that base_data is of shape (N, 2) with (N, 0) the real parts, (N,
    1) the imaginary parts
    """
    work_data = base_data.flatten(order='F')

    center = int(work_data.size / 2)
    re = work_data[0: center]
    im = work_data[center:]

    threshold = 0.1
    pha = np.abs(np.arctan2(im, re)) * 1000
    indices = np.where(pha > threshold)[0]
    if indices.size == 0:
        raise Exception(
            'Data weighting is not possible, all phase values ' +
            'below {0} mrad'.format(threshold)
        )

    re_mean = np.mean(np.abs(re))

    factor_im = re_mean / im[indices]
    errors = np.ones(work_data.shape)

    errors[center + indices] *= factor_im

    errors = errors.reshape(base_data.shape, order='F')
    return errors


def get_weighting_all_to_one(base_data):
    """
    Return a weighting vector which 'normalizes' all data points to one
    """
    errors = 1.0 / base_data
    return errors


def get_weighting_rel_abs(base_data):
    # data are in log
    # data = 10 ** base_data
    data = base_data
    errors = 0.03 * data + 0.01
    return 1.0 / errors


def get_weighting_one(base_data):
    return np.ones(base_data.shape)


functions = {
    're_vs_im': get_weighting_re_vs_im,
    'avg_im': get_weighting_im_to_avg_re,
    'avg_im_err': get_weighting_im_to_avg_re_error,
    'one': get_weighting_one,
    'all_to_one': get_weighting_all_to_one,
    'rel_abs_error': get_weighting_rel_abs,
}



"""
initial :math:`\lambda` value :math:`\lambda_0`
-----------------------------------------------

Multiple methods can be used to determine a suitable :math:`\lambda_0` value:

* (recommended) The mean of the diagonal entries of
  :math:`\underline{\underline{A}}^T \underline{\underline{W}}_d^T
  \underline{\underline{W}}_d \underline{\underline{A}}`

* *easylam*: The number of model parameters

* user determined

lambda selection for each iteration
-----------------------------------

* Fixed

* Factor/Quotient method: The lambda value of the last iteration is multiplied
  by 10 and 100; and divided by 10 and 100. The lambda values corresponding to
  the smallest resulting RMS-value is used for the next iteration.

Not implemented (yet):

* Golden section / Binary search

* Univariate search

* Line search

* L-curve (not implemented as an automatic :math:`\lambda` selector

"""
import sys
from plot_helper import *
import scipy.sparse as sparse
import numpy as np
import itertools


class Lam0_Fixed(object):
    def __init__(self, fixed_lambda):
        self.value = fixed_lambda

    def get(self, it):
        return self.value


class Lam0_Easylam(object):
    def get(self, it):
        """
        A reasonable initial value for lambda is the number of parameters.
        """
        nr_of_parameters = it.Model.obj.get_data_base_size()
        return nr_of_parameters


class Lam0_AtWtWA(object):
    def get(self, it):
        """
        Determine the initial lambda value based on

        - Newman and Alaumbaugh, 1997
        - Kemna, 2000 (p.76)

        by using the mean value of the diagonal entries of A^T W_d^T D_d A
        """
        Wd = sparse.csc_matrix(self.Data.Wd)
        J = it.Model.J(it.m)
        J = sparse.csc_matrix(J)

        M = J.T.dot(Wd.T.dot(Wd.dot(J)))
        lam0 = np.mean(np.diag(M))
        return lam0


class BaseLambda(object):
    """
    Base class for lambda (regularization parameter) function. Here we
    implement the check for lambda 0

    """
    def get_lambda(self, it, WtWm, lam_index):
        """
        Parameters
        ----------
        it : iteration for which the model update will be computed
        WtWm : the regularization matrix for which the lambda will be computed
               for
        lam_index : the index in the iteration.lams list. Used to get the
                    lambda of the last iteration

        """
        if(it.nr == 0):
            lam = self.lam0_obj.get(it)
        else:
            old_lam = it.lams[lam_index]
            lam = self._get_lambda(it, WtWm, old_lam)
        return lam


class Lcurve(object):
    """
    Create an L-curve for a given iteration.

    WARNING: This is not yet a finalized regulariazion object for use as an
    regularization object. For now it is used to plot the l-curve for the given
    iteration.
    """
    def __init__(self):
        # which rms key to use for optimization
        self.rms_key = 'rms_re_im_noerr'
        self.rms_index = 1

    def plot_lcurve(self, it, WtWms, lams, lam_index, output_prefix):
        rms_values, test_Rm, test_lams = self._sample_lambdas(it, WtWms, lams,
                                                              lam_index)
        print('rms_values', rms_values)
        print('test_Rm', test_Rm)
        print(test_lams)
        if(test_lams == []):
            print('No test lambdas available, returning')
            return

        fig, ax = plt.subplots(1, 1)
        ax.loglog(rms_values, test_Rm, '.-', color='k')
        index = 0
        for x, y in zip(rms_values, test_Rm):
            ax.annotate('{0:.4}'.format(float(test_lams[index])), xy=(x, y))
            index += 1

        ax.set_xlim([min(rms_values), max(rms_values)])
        ax.set_ylim([min(test_Rm), max(test_Rm)])
        ax.set_xlabel('RMS value: {0}'.format(self.rms_key.replace('_', '\_')))
        ax.set_ylabel(r'$\left| R \cdot m \right|$')
        ax.set_title('L-Curve based on iteration {0}'.format(it.nr))
        filename = output_prefix + 'l-curve-nr_{0}'.format(it.nr)
        fig.savefig(filename + '.png')

        # save data to text files
        header = '# lambda Rm RMS\n'
        output_data = np.vstack((test_lams, test_Rm, rms_values)).T
        with open(filename + '.dat', 'w') as fid:
            fid.write(header)
            np.savetxt(fid, output_data)

    def _sample_lambdas(self, it, WtWms, lams, lam_index):
        # these are the lambda values we will test
        test_lams_raw = np.logspace(-3, 3, 20)

        test_its = []
        test_lams = []
        test_Rm = []

        # here we store the lambda of all regularisation functions we use
        # we only change one (with index lam_index) for checks in the
        # l-curve
        lam_set = list(lams)

        # create iteration for each test lambda
        for test_lam in test_lams_raw:
            try:
                # try to create an iteration for this lambda
                print('Test lambda: ', test_lam)
                test_it = it.copy()
                lam_set[lam_index] = test_lam
                # update_m = test_it._model_update((test_lam, ), (WtWm, ))
                update_m = test_it._model_update(lam_set, WtWms)

                # we could select an optimal alpha value here
                # alpha = self.Model.steplength_selector.get_steplength(
                #   self, update_m)
                alpha = 1
                test_it.m = it.m + alpha * update_m
                test_it.f = test_it.Model.f(test_it.m)
                Rm = np.sum(WtWms[lam_index].dot(test_it.m) ** 2)
                test_Rm.append(Rm)
                test_its.append(test_it)
                test_lams.append(test_lam)
            except Exception, e:
                print('There was an error in the lambda test for ' +
                      'lambda: {0}. Trying next lambda.'.format(test_lam))
                print(e)
                continue

        # aggregate all RMS values
        rms_values = [x.rms_values[self.rms_key][self.rms_index]
                      for x in test_its]

        return rms_values, test_Rm, test_lams

    def _get_lambda(self, it, WtWm, lam_index):
        """
        this function call is compatible to the other regularization parameter
        selection functions, although not yet implemented to return an optimal
        lambda value.
        """
        # rms_values, test_Rm, test_lams = self._sample_lambdas(it, WtWm,
        #                                                      lam_index)
        # now find the optimal lambda values and return it
        return None


class SearchLambdaIndividual(BaseLambda):
    """
    This works only for dimension 0 regularizations!

    Test multiple lambda values for an optimal rms decrease

    Test for each spectrum individually and return a sparse diagonal matrix
    """
    def __init__(self, lam0_obj):
        """
        Parameters
        ----------
        lam0_obj : object which generates the first lambda values
        """
        self.lam0_obj = lam0_obj
        # which rms key to use for optimization
        self.rms_key = 'rms_re_im_noerr'
        self.rms_index = 1

    def _get_lambda(self, it, WtWm, old_lam):
        M = it.Model.convert_to_M(it.m)

        # the lambda search object we use to determine the individual lambdas
        SL = SearchLambda(self.lam0_obj)

        # extract the regularization of this spectrum from the large matrix
        len_m_small = it.Model.M_base_dims[0][1]
        WtWm_small = WtWm[0:len_m_small, 0:len_m_small]

        # loop through all spectra in the order we would flatten them
        # this is done by creating indices for all extra dimensions
        e_indices = [range(0, x[1][1]) for x in it.Data.extra_dims.items()]
        lambdas = []

        for nr, extra_index in enumerate(itertools.product(*e_indices)):
            print 'index', extra_index

            # create an iteration with only this spectrum
            it_indiv = it.copy()
            it_indiv.Data.extra_mask = list(extra_index)
            # we assume that there is only one base dimension in M
            m_slice = [slice(0, M.shape[0]), ] + list(extra_index)
            it_indiv.m = M[tuple(m_slice)]
            it_indiv.f = it.Model.f(it_indiv.m)

            # # find optimal lambda for this iteration
            if(type(old_lam) is float or type(old_lam) is int):
                lam_old_individual = old_lam
            else:
                # extract from matrix
                # compute first index
                lam_index = len_m_small * nr
                lam_old_individual = old_lam[lam_index, lam_index]

            new_lam = SL._get_lambda(it_indiv, WtWm_small, lam_old_individual)

            # add N lambdas to list (N = number parameters for this spectrum)
            lambdas += it_indiv.m.size * [new_lam, ]
        # reset the extra_mask
        it.Data.extra_mask = None
        L = sparse.csc_matrix(np.diag(lambdas))
        return L


class SearchLambda(BaseLambda):
    """
    Test multiple lambda values for an optimal rms decrease
    """
    def __init__(self, lam0_obj, rms_key='rms_all_noerr', rms_index=0):
        """
        Parameters
        ----------
        lam0_obj : object which generates the first lambda values
        """
        self.lam0_obj = lam0_obj
        # which rms key to use for optimization
        self.rms_key = rms_key
        self.rms_index = rms_index

    def _get_lambda(self, it, WtWm, old_lam):
        """
        For the actual lambda value, compute model updates for lambda / 10;
        lambda * 10 and choose the version with the best rms

        Parameters
        ----------
        see BaseLambda.get_lambda(...)
        """
        lam_old = float(old_lam)

        # these are the lambda values we will test
        # for the frozen data sets we sometimes needed:
        # lam_old/1e7, lam_old/1e4
        # lam_old / 200, lam_old / 100, lam_old / 50,
        test_lams_raw = (lam_old / 10, lam_old / 5, lam_old * 5,
                         lam_old * 10, lam_old * 100, lam_old * 1e4)

        test_its = []
        test_lams = []

        # add initial iteration and old lambda
        test_its.append(it)
        test_lams.append(lam_old)

        # create iteration for each test lambda
        for test_lam in test_lams_raw:
            try:
                # try to create an iteration for this lambda
                print('Test lambda: ', test_lam)
                test_it = it.copy()
                update_m = test_it._model_update((test_lam, ), (WtWm, ))
                # select optimal alpha value
                alpha = test_it.Model.steplength_selector.get_steplength(
                    test_it, update_m, True)
                # check if all steplengths lead to errors
                if(alpha is None):
                    continue
                # alpha = 1
                test_it.m = it.m + alpha * update_m
                test_it.f = test_it.Model.f(test_it.m)

                test_its.append(test_it)
                test_lams.append(test_lam)
            except:
                print('There was an error in the lambda test for ' +
                      'lambda: {0}. Trying next lambda.'.format(test_lam))
                e = sys.exc_info()[0]
                print e
                continue

        # aggregate all RMS values
        rms_values = [x.rms_values[self.rms_key][self.rms_index]
                      for x in test_its]

        minimal_lambda_index = np.argmin(rms_values)
        best_lam = test_lams[minimal_lambda_index]
        print('all lambdas', test_lams)
        print('optimal lambda', best_lam)

        return best_lam


class FixedLambda(BaseLambda):
    """
    Implement various regularization parameters
    """
    def __init__(self, fixed_value):
        self.fixed_lambda = fixed_value

    def get_lambda(self, it, WtWm, lam_index):
        """
        Depending on a given Iteration, return a new regularization value
        """
        return self.fixed_lambda



#!/usr/bin/env python
from setuptools import setup
import sys
# from setuptools import find_packages
# find_packages

# under windows, run
# python.exe setup.py bdist --format msi
# to create a windows installer

# TODO: understand the inclusion of requirements
version_short = '0.8'
version_long = '0.8.0'

extra = {}
if sys.version_info >= (3,):
    print('V3')
    extra['use_2to3'] = True

if __name__ == '__main__':
    setup(name='geccoinv',
          version=version_long,
          description=''.join((
              'Multi dimensional geophysical inversion framework. ',
              'Note: only for use with the ccd package. If you are looking ',
              'for a general inversion framework, try pyGimli ',
              '(www.pygimli.org)',
          )),
          author='Maximilian Weigand',
          author_email='mweigand@geo.uni-bonn.de',
          url='http://www.geo.uni-bonn.de',
          # find_packages() somehow does not work under Win7 when creating a
          # msi # installer
          # packages=find_packages(),
          package_dir={'': 'lib'},
          packages=[
              'NDimInv',
              'sip_formats',
          ],
          scripts=[],
          install_requires=['numpy', 'scipy>=0.12', 'matplotlib'],
          **extra
          )



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This file contains some very old functions to handle grids used by CRTomo (see
Kemna, 2000): "Tomographic inversion of complex resistivity -- theory and
application".

GeccoInv (and dd_interfaces) was created to (among other applications) analyze
data produced by CRTomo. Thus this file.

Note that this file does not follow any coding standard, no OOP, and makes heavy
use of global variables. Do not use it for further development. It's only here
to make some functionality work until someone finds time to reimplement the
functionality in a proper way.

Copyright 2014 Maximilian Weigand

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import numpy.ma as ma
import os
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

# grid variables:
global nodes, nodes_sorted
nodes = None
nodes_sorted = None

global grid_is_rectangular
grid_is_rectangular = True

# for triangle grids
global triangles
triangles = None

# for rectangular grids
global gridx, gridz
gridx = None
gridz = None


class cls_configs:
    def __init__(self):
        self.A = None
        self.B = None
        self.M = None
        self.N = None

configs = cls_configs()
global width_in_elements, height_in_elements
width_in_elements = -1
height_in_elements = -1

global electrode_pos_x, electrode_pox_z
electrode_pos_x = None
electrode_pos_z = None


class element:
    def __init__(self):
        self.nodes = np.empty(1)
        self.xcoords = np.empty(1)
        self.zcoords = np.empty(1)

global element_type_list
element_type_list = []

global filter_mask
filter_mask = None

"""
The following lists hold the grid data that will be plotted
"""
global element_data
element_data = None
global config
config = None

global node_data
node_data = None

"""
Plot options
"""


class plot_options:
    def __init__(self):
        self.cbmin = None
        self.cbmax = None
        self.grayscale = False
        self.reverse = False
        self.cblabel = ''
        self.title = ''
        self.title_position = None
        self.xlabel = 'x (cm)'
        self.ylabel = 'z (cm)'
        self.xmin = None
        self.xmax = None
        self.zmin = None
        self.zmax = None
        self.norm_fac = None
        self.dyn_fac = None
        self.cb_nr_tiks = None
        self.x_nr_tiks = None
        self.y_nr_tiks = None
        self.clip_electrodes = False
        self.no_elecs = False


global plt_opt
plt_opt = plot_options()

# global figure for plotting
global fig
fig = plt.figure()


def get_colormap():
    """
    Depending on the user settings, return a colormap
    """
    if(plt_opt.grayscale is False):
        if(plt_opt.reverse is True):
            cmap = mpl.cm.jet
        else:
            cmap = mpl.cm.jet_r
    else:
        if(plt_opt.reverse is True):
            cmap = mpl.cm.binary_r
        else:
            cmap = mpl.cm.binary
    return cmap


def plot_elements_to_file(output, cid, scale='linear'):
    """
    Plot element data to a grid and save to file
    """
    fig.clf()
    plot_element_data(cid, 111, scale)
    plt.savefig(output, dpi=300, bbox_inches='tight')


def plot_nodes_to_file(output, cid, fill_contours=True, levels=None):
    """
    Plot node data as a contour plot on a grid and save to file
    """
    fig.clf()
    plot_node_data(cid, 111, fill_contours, levels)
    plt.savefig(output, dpi=300, bbox_inches='tight')


def plot_element_data(cid, subplot_id, scale='linear'):
    """
    Plot element data to a grid. Only plot to the global figure,
    no output.
    """
    # frame_on controls the frame
    ax = fig.add_subplot(subplot_id, frame_on=False)
    if(plt_opt.no_elecs):
        no_elecs = True
    else:
        no_elecs = False
    plot_element_data_to_ax(cid, ax, scale, no_electrodes=no_elecs)
    return ax


def plot_node_data(cid, subplot_id, fill_contours=True, levels=None):
    """
    Plot node data to a grid as a contour plot. Only plot to the global figure,
    no output.
    """
    mpl.rcParams["font.size"] = 24.0

    x, y = get_dimensions_in_elements()
    X, Y = get_grid()

    xlim, zlim = _get_grid_limits()
    colors = np.reshape(node_data[:, cid], (y + 1, x + 1), order='C')

    # frame_on controls the frame
    ax = fig.add_subplot(subplot_id, frame_on=False)

    cmap = get_colormap()
    if(plt_opt.cbmin is None):
        plt_opt.cbmin = np.min(colors)
    if(plt_opt.cbmax is None):
        plt_opt.cbmax = np.max(colors)

    norm = mpl.colors.Normalize(vmin=float(plt_opt.cbmin),
                                vmax=float(plt_opt.cbmax))

    # levels for contourf
    if(levels is None):
        levels_neg = -np.logspace(-2, np.log10(np.abs(np.min(colors))), 14)
        levels_pos = np.logspace(-2, np.log10(np.max(colors)), 14)
        levels = np.sort(np.hstack((levels_neg, levels_pos)))

    if(fill_contours):
        pm = ax.contourf(X, Y, colors, levels=levels, cmap=cmap, norm=norm)
    else:
        pm = ax.contour(X, Y, colors, levels=levels, cmap=cmap, norm=norm)

    # version check
    if(int(mpl.__version__[0]) == 1):
        print('good version')
        # colorbar same hight as plot
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size=0.12, pad=0.15)
        cb = plt.colorbar(pm, orientation='vertical', cax=cax)
    else:
        print('old version')
        cb = plt.colorbar(pm, orientation='vertical')

    # labels
    # cb.set_label(options.cblabel)

    ax.set_xlabel(plt_opt.xlabel)
    ax.set_ylabel(plt_opt.ylabel)
    ax.set_title(plt_opt.title)

    ax.titleOffsetTrans._t = (0., 10.0 / 72.0)
    # default ist (0., 5/72)
    ax.titleOffsetTrans.invalidate()

    # bad version check
    if(mpl.__version__[0] == 1):
        # fit axes to plot
        ax.autoscale(enable=True, axis='both', tight=True)
        # ax.autoscale_view()  # tight scaling to image extent

    # plot electrodes
    plot_electrodes_to_ax(ax)

    ax.set_xlim(xlim)
    ax.set_ylim(zlim)
    ax.set_aspect('equal')

    # run through all lines drawn for xticks and yticks
    for i, line in enumerate(ax.get_xticklines() + ax.get_yticklines()):
        line.set_visible(False)

    return ax, cb


def _get_grid_limits():
    """
    Depending on the user settings and the node coordinates, return the grid
    limits to use for plotting.
    """
    nodes = get_nodes()

    xlim = [0, 0]
    if(plt_opt.xmin is None or plt_opt.xmin == '*'):
        xlim[0] = np.min(nodes[:, 1])
    else:
        xlim[0] = float(plt_opt.xmin)

    if(plt_opt.xmax is None or plt_opt.xmax == '*'):
        xlim[1] = np.max(nodes[:, 1])
    else:
        xlim[1] = float(plt_opt.xmax)

    zlim = [0, 0]
    if(plt_opt.zmin is None or plt_opt.zmin == '*'):
        zlim[0] = np.min(nodes[:, 2])
    else:
        zlim[0] = float(plt_opt.zmin)

    if(plt_opt.zmax is None or plt_opt.zmax == '*'):
        zlim[1] = np.max(nodes[:, 2])
    else:
        zlim[1] = float(plt_opt.zmax)

    return xlim, zlim


def _get_colors(cid, scale):
    """
    Return the color values for the grid cells, and the limits (min, max).

    Also, transform the data according to the required scale.
    Possible scale values are:
    'linear' - do not transform the data at all
    'log' - log10 the data
    '10' - 10 ** the data
    'asinh' - perform a arcsinh transformation on the data. Consult the CRLab
              manual for detailed information (section about sens.dat).
    """
    if(grid_is_rectangular):
        x, y = get_dimensions_in_elements()
        colors = np.reshape(element_data[:, cid], (y, x), order='C')
    else:
        colors = np.squeeze(element_data[:, cid])

    if(scale == 'log' or scale == 'log10'):
        colors = np.log10(colors)
    elif(scale == '10'):
        colors = 10 ** colors
    elif(scale == 'asinh'):
        if(plt_opt.norm_fac is None):
            norm_fac = np.max(np.abs(colors))
        else:
            norm_fac = plt_opt.norm_fac

        if(plt_opt.dyn_fac is None):
            dyn_fac = np.abs(np.min(np.log10(np.abs(colors))))
        else:
            dyn_fac = plt_opt.dyn_fac

        colors = np.arcsinh(10 ** (dyn_fac) * colors / norm_fac) /\
            np.arcsinh(10 ** dyn_fac)

        if(plt_opt.cblabel == 'fill'):
            plt_opt.cblabel = 'norm = {0}'.format(norm_fac)
            plt_opt.title = r'$sinh^{-1}(10^{' + '{0:.3f}'.format(dyn_fac) +\
                            '} \cdot S_{norm}) / sinh^{-1}(10^{' +\
                            '{0:.3f}'.format(dyn_fac) + '})$'

    colors = ma.masked_where(np.isnan(colors), colors)
    colors = ma.masked_where(np.isinf(colors), colors)

    if(filter_mask is not None):
        print('Applying filter mask')
        if(grid_is_rectangular):
            colors = ma.masked_where(np.reshape(filter_mask, (y, x),
                                                order='C') == 0, colors)
        else:
            colors = ma.masked_where(filter_mask == 0, colors)

    if(plt_opt.cbmin is None):
        cbmin = np.min(colors)
    else:
        cbmin = plt_opt.cbmin

    if(plt_opt.cbmax is None):
        cbmax = np.max(colors)
    else:
        cbmax = plt_opt.cbmax

    if(cbmax < cbmin):
        cbmin, cbmax = cbmax, cbmin

    # if the data has only one value, adjust the cb-limits
    if(cbmin == cbmax):
        cbmin = cbmin - 0.2
        cbmax = cbmax + 0.2

    return colors, cbmin, cbmax


def plot_element_data_to_ax(cid, ax, scale='linear', no_cb=False,
                            no_electrodes=False, plot_gridlines=False):
    """
    Plot element to to a grid, given an axis object

    Parameters
    ----------
    cid : id for previously loaded data
    ax : axes object to plot to
    scale : [linear|log10] (see _get_colors for more options)
    no_cb : plot colorbar or not
    plot_gridlines : plot the grid
    """
    xlim, zlim = _get_grid_limits()

    colors, cbmin, cbmax = _get_colors(cid, scale)

    # we want to export those
    global cmap
    global norm
    cmap = get_colormap()
    norm = mpl.colors.Normalize(vmin=float(cbmin), vmax=float(cbmax))

    if(plot_gridlines):
        edgecolor = 'k'
    else:
        edgecolor = 'face'

    if(grid_is_rectangular):
        X, Y = get_grid()
        pm = ax.pcolormesh(X, Y, colors, edgecolor=edgecolor, linewidth=1e-2,
                           cmap=cmap, norm=norm, rasterized=True)
    else:
        tx = nodes_sorted[:, 1]
        ty = nodes_sorted[:, 2]
        # edgecolor='k' plots the grid lines
        pm = ax.tripcolor(tx, ty, triangles, edgecolor=edgecolor,
                          facecolors=colors, cmap=cmap, norm=norm)

    if(no_cb is False):
        # version check
        if(int(mpl.__version__[0]) == 1):
            # colorbar same hight as plot
            from mpl_toolkits.axes_grid1 import make_axes_locatable
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size=0.12, pad=0.15)
            cb = plt.colorbar(pm, orientation='vertical', cax=cax)
        else:
            print('You are using an old version of Matplotlib ({0}).' +
                  'The colorbar can only be resized according to some crude' +
                  'heuristics!'.format(mpl.__version__))
            # determine X and Z dimensions
            width = np.abs(xlim[1] - xlim[0])
            height = np.abs(zlim[1] - zlim[0])
            if(height < width):
                ratio = height / width
            else:
                ratio = width / height

            cax, kw = mpl.colorbar.make_axes(ax, shrink=ratio)
            cb = plt.colorbar(pm, cax=cax, orientation='vertical')

        if(plt_opt.cb_nr_tiks is not None):
            cb.locator = mpl.ticker.MaxNLocator(nbins=plt_opt.cb_nr_tiks)
            cb.update_ticks()

        # labels
        cb.set_label(plt_opt.cblabel)
    else:
        cb = None
    ax.set_xlabel(plt_opt.xlabel)
    ax.set_ylabel(plt_opt.ylabel)

    if (plt_opt.title_position is None):
        ax.set_title(plt_opt.title)
    else:
        ax.set_title(plt_opt.title, position=plt_opt.title_position)

    # ax.title.set_y(1.05) # check way to set the title padding
    ax.titleOffsetTrans._t = (0., 10.0 / 72.0)
    # default ist (0., 5/72)
    ax.titleOffsetTrans.invalidate()

    # bad version check
    if(int(mpl.__version__[0]) == 1):
        # fit axes to plot
        ax.autoscale(enable=True, axis='both', tight=True)
        # ax.autoscale_view()  # tight scaling to image extent

    # plot electrodes
    if(not no_electrodes):
        plot_electrodes_to_ax(ax)

    # why do we need this?
    # fig.tight_layout()

    # run through all lines drawn for xticks and yticks
    for i, line in enumerate(ax.get_xticklines() + ax.get_yticklines()):
        line.set_visible(False)

    # set user supplied nr of x/y tick labels
    if(plt_opt.x_nr_tiks is not None):
        ax.xaxis.set_major_locator(mpl.ticker.MaxNLocator(
            nbins=plt_opt.x_nr_tiks))

    if(plt_opt.y_nr_tiks is not None):
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(
            nbins=plt_opt.y_nr_tiks))

    ax.set_xlim(xlim)
    ax.set_ylim(zlim)
    ax.set_aspect('equal')

    # ax.get_figure().canvas.draw()  # needed to get the labels/positions
    # # now we want to check for overlapping y tick labels
    # axy = ax.get_yaxis()
    # #print dir(axy)
    # bboxes = []
    # for i in axy.get_majorticklabels():
    #    #print i.get_label()
    #    #print i.get_text()
    #    #print dir(i)
    #    bboxes.append(i._bbox)
    # print bboxes

    return ax, pm, cb


def plot_electrodes_to_ax(ax, add_numbers=False):
    """
    Plot the electrodes on a given axes object.

    If add_numbers == True, then annotate each electrode with its number
    """
    Ex, Ez = get_electrodes()

    nr = 1
    for x, z in zip(Ex, Ez):
        if((plt_opt.xmin == '*' or plt_opt.xmin is None or
            x >= float(plt_opt.xmin)) and
           (plt_opt.xmax == '*' or plt_opt.xmax is None or
            x <= float(plt_opt.xmax)) and
           (plt_opt.zmin == '*' or plt_opt.zmin is None or
            z >= float(plt_opt.zmin)) and
           (plt_opt.zmax == '*' or plt_opt.zmax is None or
                z <= float(plt_opt.zmax))):
            ax.scatter(x, z, c='k', s=10, clip_on=plt_opt.clip_electrodes)

        if(add_numbers):
            ax.annotate('{0}'.format(nr), (x, z), fontsize=5.0,
                        textcoords='offset points', xytext=(3, 3))
        nr += 1


def plot_electrodes_to_ax_selected(ax, add_numbers=False, add_labels=None,
                                   selected_electrodes=None):
    """
    Plot the electrodes on a given axes object.

    If add_numbers == True, then annotate each electrode with its number
    """
    Ex, Ez = get_electrodes(selected_electrodes)

    nr = 1
    for x, z in zip(Ex, Ez):
        if((plt_opt.xmin == '*' or plt_opt.xmin is None or
            x >= float(plt_opt.xmin)) and
            (plt_opt.xmax == '*' or plt_opt.xmax is None or
             x <= float(plt_opt.xmax)) and
            (plt_opt.zmin == '*' or plt_opt.zmin is None or
             z >= float(plt_opt.zmin)) and
            (plt_opt.zmax == '*' or plt_opt.zmax is None or
                z <= float(plt_opt.zmax))):
            ax.scatter(x, z, c='k', s=10, clip_on=False)

        if(add_numbers):
            ax.annotate('{0}'.format(nr), (x, z), fontsize=5.0,
                        textcoords='offset points', xytext=(3, 3))
        nr += 1


def load_column_file_to_elements(filename, columns):
    """
    Import data to be plotted on the grid from a file with the following
    structure:

    1. Line: Number of entries
    2. - to [Number of entries + 1]: One or more columns of data points

    Parameters
    ----------
    filename : file to import
    columns : list with indices to load, starting with 0

    WARNING: This function assumes that the first line contains as many columns
    as the rest of the file!

    Return a list with the entries in the global data matrix, which were added

    """
    tmp_data = np.loadtxt(filename, dtype=float, skiprows=1)

    global element_data
    if(element_data is None):
        element_data = tmp_data[:, columns]
        indices = range(0, element_data.shape[1])
    else:
        indices = element_data.shape[1]
        element_data = np.append(element_data, tmp_data[:, columns], axis=1)
        indices = range(indices, element_data.shape[1])
    return indices


def load_column_file_to_elements_advanced(
    filename, columns, first_line_contains_nr_elements=True,
        no_header_line=False):
    """
    Import column data from a data file.

    Read out the first line to get the number of elements (rows).
    Otherwise behave as load_column_file_to_elements

    Parameters
    ----------

    filename: Data filename containing column data
    columns: List of column ids to import, starting from zero. Examples:
             [0,4,3] or [1,]

    first_line_contains_nr_elements: False - read as many elements a loaded in
                                     the grid. Exit if no grid was
                                     loaded yet!
                                     True  - read as many elements as written in
                                     the first line

    no_header_line: True: No header line is expected, the first line is assumed
                    to be data

    Returns
    -------
    indices: list of internal indices that identify the loaded columns. These
             indices can the be passed to the plotting functions.

    """

    fid = open(filename, 'r')

    if(first_line_contains_nr_elements is True):
        first_line = fid.readline().strip()
        first_line = np.fromstring(first_line, dtype=int, sep=' ')[0]
    else:
        first_line = len(element_type_list[0])
        if(no_header_line is False):
            fid.readline()
    tmp_data = np.zeros((first_line, len(columns)))
    for line_nr in range(0, first_line):
        line_str = fid.readline().strip()
        tmp_data[line_nr, :] = np.fromstring(line_str, dtype=float,
                                             sep=' ')[columns]

    indices = add_to_element_data(tmp_data)

    return indices


def add_to_element_data(subdata, fill_or_cut=False):
    """
    Add a new element data set to the global list of data sets. Return the
    index to this new data set.

    Parameters
    ----------
    element_data has two dimensions: [nr_of_cells, ids]
    fill_or_cut: if True, then fill with NaN values or cut to corresponding
                 element nr
    """
    # make sure we have a two dimension subdata
    if len(subdata.shape) == 1:
        subdata_2d = np.atleast_2d(subdata).T
    else:
        subdata_2d = subdata

    nr_elements = get_nr_of_elements()
    if fill_or_cut and subdata_2d.shape[0] > nr_elements:
        subdata_2d = subdata_2d[0:nr_elements, :]
    if fill_or_cut and subdata_2d.shape[0] < nr_elements:
        tmp = np.ones((nr_elements, subdata_2d.shape[1])) * np.nan
        tmp[0:subdata_2d.shape[0], :] = subdata_2d
        subdata_2d = tmp

    global element_data
    if element_data is None:
        element_data = subdata_2d
        indices = range(0, element_data.shape[1])
    else:
        indices = element_data.shape[1]
        element_data = np.append(element_data, subdata_2d, axis=1)
        indices = range(indices, element_data.shape[1])
    return indices


def load_column_file_to_nodes(filename, columns):
    """
    Important: For node files there is no first line with the number of
    elements
    """
    tmp_data = np.loadtxt(filename, dtype=float)
    tmp_data[:, 0:2] = np.round(tmp_data[:, 0:2], 5)

    tmp_data = resort_node_data(tmp_data)

    global node_data
    if(node_data is None):
        node_data = tmp_data[:, columns]
        indices = range(0, node_data.shape[1])
    else:
        indices = node_data.shape[1]
        node_data = np.append(node_data, tmp_data[:, columns], axis=1)
        indices = range(indices, node_data.shape[1])
    return indices


def convert_to_conductivity(mid, pid):
    """
    For given element ids, assume resistivity magnitude and phase. Then,
    convert to conductivity and return the conductivity magnitude and phase ids
    of the element_data array.

    We assume that the magnitudes are in log10!
    """
    global element_data

    # get data
    magnitudes = 10 ** element_data[:, mid]
    phases = element_data[:, pid]

    # convert
    resistivities = magnitudes * np.exp(1j / 1000 * phases)
    conductivities = 1 / resistivities
    mag_cond = np.abs(conductivities)
    pha_cond = np.arctan(np.imag(conductivities) / np.real(conductivities))
    # convert to mrad
    pha_cond *= 1000

    # save
    indices = [element_data.shape[1], element_data.shape[1] + 1]
    element_data = np.append(element_data, mag_cond, axis=1)
    element_data = np.append(element_data, pha_cond, axis=1)

    return indices


def convert_to_conductivity_re_im(mid, pid):
    """
    For given element ids, assume resistivity magnitude and phase. Then,
    convert to conductivity and return the conductivity magnitude and phase ids
    of the element_data array.

    We assume that the magnitudes are in log10!
    """
    global element_data

    # get data
    magnitudes = 10 ** element_data[:, mid]
    phases = element_data[:, pid]

    # convert
    impedances = magnitudes * np.exp(1j / 1000 * -phases)
    conductance = 1 / impedances
    real_cond = np.real(conductance)
    imag_cond = np.imag(conductance)

    # save
    indices = [element_data.shape[1], element_data.shape[1] + 1]
    element_data = np.append(element_data, real_cond, axis=1)
    element_data = np.append(element_data, imag_cond, axis=1)

    return indices


def resort_node_data(node_data):
    """
    If we read node data from a file, we have to take acount the changed
    positions due to CutMcK.
    """

    X, Y = get_grid()
    tpots = np.empty(X.shape)
    for rows in range(0, X.shape[0]):
        for cols in range(0, X.shape[1]):
            a = np.where(node_data[:, 0] == X[rows, cols])
            b = np.where(node_data[:, 1] == Y[rows, cols])
            c = np.intersect1d(a[0], b[0])
            # resort column index 2
            tpots[rows, cols] = node_data[c, 2]

    node_data[:, 2] = tpots.reshape(node_data.shape[0])
    return node_data


def load_pot_file(filename):
    """
    Wrapper for load_column_file_to_nodes to load a pot.dat file, i.e. columns
    3 and 4. Return the indices.
    """
    indices = load_column_file_to_nodes(filename, [2, 3])
    return indices


def load_mag_file(filename):
    index = load_column_file_to_elements(filename, [2, ])
    return index


def load_pha_file(filename):
    index = load_column_file_to_elements(filename, [2, ])
    return index


def load_elem_file(filename):
    """
    Determine the type (regular, irregular) of grid we are dealing with an call
    the corresponding load function.
    """
    if(not os.path.isfile(filename)):
        print('Grid file "{0}" not found'.format(filename))
        exit()
    if(nodes is not None):
        # grid already loaded
        return
    # TODO: do something here and implement irregular grids
    load_regular_elem_file(filename)


def load_regular_elem_file(filename):
    """
    Load an elem.dat file containing a regular grid, process it, and store
    the data in a global variable
    """
    global width_in_nodes
    global height_in_nodes
    f = open(filename)
    # first header line
    firstline = f.readline().lstrip()
    number_of_nodes, number_of_element_types, bandwidth = np.fromstring(
        firstline, dtype=int, sep='   ')

    # read header lines for each element type
    element_infos = np.zeros((number_of_element_types, 3), dtype=int)
    for element_type in range(0, number_of_element_types):
        element_line = f.readline().lstrip()
        element_infos[element_type] = np.fromstring(element_line, dtype=int,
                                                    sep='  ')

    # prepare nodes
    global nodes, nodes_sorted
    nodes_sorted = np.zeros((number_of_nodes, 3), dtype=float)
    nodes = np.zeros((number_of_nodes, 3), dtype=float)

    # read in nodes
    for node in range(0, number_of_nodes):
        node_line = f.readline().lstrip()
        nodes[node, :] = np.fromstring(node_line, dtype=float, sep='    ')

    # Rearrange nodes when CutMcK was used.
    # The check is based on the first node, but if one node was renumbered, so
    # were all the others.
    global nodes_cutmck_index
    nodes_cutmck_index = None
    if(nodes[0, 0] != 1 or nodes[1, 0] != 2):
        nodes_cutmck_index = np.zeros(nodes.shape[0])
        for node in range(0, number_of_nodes):
            new_index = np.where(nodes[:, 0].astype(int) == (node + 1))
            nodes_sorted[new_index[0], 1:3] = nodes[node, 1:3]
            nodes_sorted[new_index[0], 0] = new_index[0]
            nodes_cutmck_index[node] = new_index[0]
        print('This grid was sorted using CutMcK. The nodes were resorted!')
    else:
        nodes_sorted = nodes

    # round node coordinates to 5th decimal point. Sometimes this is important
    # when we deal with mal-formatted node data
    nodes_sorted[:, 1:3] = np.round(nodes_sorted[:, 1:3], 5)

    global element_type_list

    # read elements
    for element_type in range(0, number_of_element_types):
        element_list = []
        for element_coordinates in range(0,
                                         int(element_infos[element_type, 1])):
            element_coordinates_line = f.readline().lstrip()
            tmp_element = element()
            tmp_element.nodes = np.fromstring(element_coordinates_line,
                                              dtype=int, sep=' ')
            tmp_element.xcoords = nodes_sorted[tmp_element.nodes - 1, 1]
            tmp_element.zcoords = nodes_sorted[tmp_element.nodes - 1, 2]
            element_list.append(tmp_element)
        element_type_list.append(element_list)

    # depending on the type of grid (rectangular or triangle), prepare grids or
    # triangle lists
    global grid_is_rectangular
    if(len(element_type_list[0][0].nodes) == 3):
        print('Triangular grid found')
        grid_is_rectangular = False

        # prepare triangles
        global triangles
        triangles = element_type_list[0]
        triangles = [x.nodes for x in triangles]
        # python starts arrays with 0, but elem.dat with 1
        triangles = np.array(triangles) - 1

    else:
        print('Rectangular grid found')
        grid_is_rectangular = True
        # calculate the dimensions of the grid
        calculate_dimensions(nodes)

        # ok, we kind of ignore the actual elements and just create a regular
        # grid using the nodes.
        # TODO: This is a no-go if we ever encounter rectangular grids with
        #       topography
        xcoordinates = nodes[range(0, width_in_nodes), 1]
        ycoordinates = nodes[range(
            0,
            int(width_in_nodes * height_in_nodes),
            width_in_nodes),
            2
        ]

        meshx, meshy = np.meshgrid(xcoordinates, ycoordinates)

        # store in global variables
        global gridx
        global gridz
        gridx = meshx
        gridz = meshy


def calc_element_centroids():
    global elm_centroids
    elm_centroids = []

    # loop over element 8
    for element in element_type_list[0]:
        A = 0
        for i in range(0, 4):
            ind1 = i
            ind2 = (i + 1) % 4
            contrib = (element.xcoords[ind1] * element.zcoords[ind2] -
                       element.xcoords[ind2] * element.zcoords[ind1])
            A += contrib
        A *= 0.5

        x_centr = 0
        for i in range(0, 4):
            ind1 = i
            ind2 = (i + 1) % 4
            contrib = (element.xcoords[ind1] + element.xcoords[ind2]) *\
                      (element.xcoords[ind1] * element.zcoords[ind2] -
                       element.xcoords[ind2] * element.zcoords[ind1])
            x_centr += contrib
        x_centr *= 1 / (6 * A)

        z_centr = 0
        for i in range(0, 4):
            ind1 = i
            ind2 = (i + 1) % 4
            contrib = (element.zcoords[ind1] + element.zcoords[ind2]) * \
            (element.xcoords[ind1] * element.zcoords[ind2] - \
                    element.xcoords[ind2] * element.zcoords[ind1])
            z_centr += contrib
        z_centr *= 1 / (6 * A)

        elm_centroids.append((x_centr, z_centr))


def calculate_dimensions(nodes):
    """
    For a regular grid (rectangular cells), compute the dimensions in x and z
    directions (# nodes, # elements)
    """
    # define global variables
    global width_in_nodes, height_in_nodes
    global width_in_elements, height_in_elements

    # determine width and height of grid in element numbers
    width_in_nodes = 0
    initial_y_value = nodes[0, 2]
    for value in range(0, len(nodes[:, 2])):
        if(initial_y_value != nodes[value, 2]):
            break
        width_in_nodes += 1

    height_in_nodes = len(nodes) / width_in_nodes

    # store in global variables
    width_in_elements = width_in_nodes - 1
    height_in_elements = height_in_nodes - 1


def load_elec_file(filename):
    """
    Load an elec.dat file
    """
    if(not os.path.isfile(filename)):
        print('Grid electrode file "{0}" not found'.format(filename))
        exit()
    # TODO: Check if grid is already loaded
    global electrode_nodes
    electrode_nodes = np.genfromtxt(filename, dtype=int)

    global electrode_pos_x, electrode_pos_z
    electrode_pos_x = np.zeros((electrode_nodes[0]))
    electrode_pos_z = np.zeros((electrode_nodes[0]))

    # remove first line: number of electrodes
    electrode_nodes = np.delete(electrode_nodes, 0)

    for index, node in enumerate(electrode_nodes):
        electrode_pos_x[index] = nodes_sorted[node - 1, 1]
        electrode_pos_z[index] = nodes_sorted[node - 1, 2]


def get_dimensions_in_elements():
    """
    return the dimensions of the grid in element numbers, return (-1, -1) if no grid was loaded yet.
    """
    return width_in_elements, height_in_elements


def get_nr_of_elements():
    return (width_in_elements * height_in_elements)


def get_electrodes():
    """
    return electrode positions
    """
    return electrode_pos_x, electrode_pos_z


def get_grid():
    """
    Return the grid in a form usable by matplotlib functions.
    Treat the output as output from meshgrid!
    """
    return gridx, gridz


def get_nodes():
    """
    Return nodes
    """
    return nodes

